/*------------------------------------------------------------------------
 |  FILE            DatabaseMetaData.java
 |  PART OF         Sverre's Simple JDBC Driver for PostgreSQL
 |  WRITTEN BY      Sverre H. Huseby <shh@thathost.com>
 +----------------------------------------------------------------------*/

package no.shhsoft.postgresql;

import java.util.*;
import java.sql.*;

/* NOTE: this class must be instantiated before the user gets to do
 * anything with the connection, as it relies on being undisturbed
 * while extracting information from the database.  Connection
 * currently handles this by creating an instance in its
 * constructor. */

/**
 * This class provides information about the database as a whole.
 *
 * <P>Many of the methods here return lists of information in ResultSets.
 * You can use the normal ResultSet methods such as getString and getInt
 * to retrieve the data from these ResultSets.  If a given form of
 * metadata is not available, these methods should throw a SQLException.
 *
 * <P>Some of these methods take arguments that are String patterns.  These
 * arguments all have names such as fooPattern.  Within a pattern String, "%"
 * means match any substring of 0 or more characters, and "_" means match
 * any one character. Only metadata entries matching the search pattern
 * are returned. If a search pattern argument is set to a null ref, it means
 * that argument's criteria should be dropped from the search.
 *
 * <P>A SQLException will be thrown if a driver does not support a meta
 * data method.  In the case of methods that return a ResultSet,
 * either a ResultSet (which may be empty) is returned or a
 * SQLException is thrown.
 *
 * @author   Sverre H. Huseby
 *           &lt;<A HREF="mailto:shh@thathost.com">shh@thathost.com</A>&gt;
 * @version  $Id: DatabaseMetaData.java,v 1.3 2002/02/05 22:54:25 sverrehu Exp $
 */
public final class DatabaseMetaData
implements java.sql.DatabaseMetaData {
/*-----------------------------------------------------------------------+
|  PRIVATE PART                                                          |
+-----------------------------------------------------------------------*/
    private static final int MAX_IDENTIFIER_LENGTH = 32;
    private static final int MAX_ROW_SIZE = 8192;
    private static final int MAX_STATEMENT_LENGTH = 8192;
    private static final int MAX_USERNAME_LENGTH = 32; /* pg_user table */

    private Connection conn;

/*-----------------------------------------------------------------------+
|  PACKAGE LOCAL PART                                                    |
+-----------------------------------------------------------------------*/
    DatabaseMetaData(Connection conn) {
        this.conn = conn;
    }

/*-----------------------------------------------------------------------+
|  PUBLIC INTERFACE                                                      |
+-----------------------------------------------------------------------*/

    /* java.sql.DatabaseMetaData ****************************************/
    // First, a variety of minor information about the target database.

    /**
     * Can all the procedures returned by getProcedures be called by the
     * current user?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean allProceduresAreCallable()
    throws SQLException {
        return true;  /* would guess so. */
        /* TODO */
    }

    /**
     * Can all the tables returned by getTable be SELECTed by the
     * current user?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean allTablesAreSelectable()
    throws SQLException {
        return true;  /* would guess so. */
        /* TODO */
    }

    /**
     * What's the url for this database?
     *
     * @return the url or null if it can't be generated
     * @exception SQLException if a database-access error occurs.
     */
    public String getURL()
    throws SQLException {
        return conn.url;
    }

    /**
     * What's our user name as known to the database?
     *
     * @return our database user name
     * @exception SQLException if a database-access error occurs.
     */
    public String getUserName()
    throws SQLException {
        return conn.username;
    }

    /**
     * Is the database in read-only mode?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean isReadOnly()
    throws SQLException {
        return conn.isReadOnly();
    }

    /**
     * Are NULL values sorted high?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean nullsAreSortedHigh()
    throws SQLException {
        return false;
    }

    /**
     * Are NULL values sorted low?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean nullsAreSortedLow()
    throws SQLException {
        return false;
    }

    /**
     * Are NULL values sorted at the start regardless of sort order?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean nullsAreSortedAtStart()
    throws SQLException {
        return false;
    }

    /**
     * Are NULL values sorted at the end regardless of sort order?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean nullsAreSortedAtEnd()
    throws SQLException {
        return true;
    }

    /**
     * What's the name of this database product?
     *
     * @return database product name
     * @exception SQLException if a database-access error occurs.
     */
    public String getDatabaseProductName()
    throws SQLException {
        return "PostgreSQL";
    }

    /**
     * What's the version of this database product?
     *
     * @return database version
     * @exception SQLException if a database-access error occurs.
     */
    public String getDatabaseProductVersion()
    throws SQLException {
        return "6.5.3 assumed";
        /* TODO: pick the correct version from somewhere */
    }

    /**
     * What's the name of this JDBC driver?
     *
     * @return JDBC driver name
     * @exception SQLException if a database-access error occurs.
     */
    public String getDriverName()
    throws SQLException {
        return "SHHSoft PostgreSQL JDBC Driver";
    }

    /**
     * What's the version of this JDBC driver?
     *
     * @return JDBC driver version
     * @exception SQLException if a database-access error occurs.
     */
    public String getDriverVersion()
    throws SQLException {
        return Driver.DRIVER_MAJOR_VERSION + "." + Driver.DRIVER_MINOR_VERSION;
    }

    /**
     * What's this JDBC driver's major version number?
     *
     * @return JDBC driver major version
     */
    public int getDriverMajorVersion() {
        return Driver.DRIVER_MAJOR_VERSION;
    }

    /**
     * What's this JDBC driver's minor version number?
     *
     * @return JDBC driver minor version number
     */
    public int getDriverMinorVersion() {
        return Driver.DRIVER_MINOR_VERSION;
    }

    /**
     * Does the database store tables in a local file?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean usesLocalFiles()
    throws SQLException {
        return false;
    }

    /**
     * Does the database use a file for each table?
     *
     * @return true if the database uses a local file for each table
     * @exception SQLException if a database-access error occurs.
     */
    public boolean usesLocalFilePerTable()
    throws SQLException {
        return false;
    }

    /**
     * Does the database treat mixed case unquoted SQL identifiers as
     * case sensitive and as a result store them in mixed case?
     *
     * A JDBC-Compliant driver will always return false.
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsMixedCaseIdentifiers()
    throws SQLException {
        return false;
    }

    /**
     * Does the database treat mixed case unquoted SQL identifiers as
     * case insensitive and store them in upper case?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean storesUpperCaseIdentifiers()
    throws SQLException {
        return false;
    }

    /**
     * Does the database treat mixed case unquoted SQL identifiers as
     * case insensitive and store them in lower case?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean storesLowerCaseIdentifiers()
    throws SQLException {
        return true;
    }

    /**
     * Does the database treat mixed case unquoted SQL identifiers as
     * case insensitive and store them in mixed case?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean storesMixedCaseIdentifiers()
    throws SQLException {
        return false;
    }

    /**
     * Does the database treat mixed case quoted SQL identifiers as
     * case sensitive and as a result store them in mixed case?
     *
     * A JDBC-Compliant driver will always return true.
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsMixedCaseQuotedIdentifiers()
    throws SQLException {
        return true;
    }

    /**
     * Does the database treat mixed case quoted SQL identifiers as
     * case insensitive and store them in upper case?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean storesUpperCaseQuotedIdentifiers()
    throws SQLException {
        return false;
    }

    /**
     * Does the database treat mixed case quoted SQL identifiers as
     * case insensitive and store them in lower case?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean storesLowerCaseQuotedIdentifiers()
    throws SQLException {
        return false;
    }

    /**
     * Does the database treat mixed case quoted SQL identifiers as
     * case insensitive and store them in mixed case?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean storesMixedCaseQuotedIdentifiers()
    throws SQLException {
        return false;
    }

    /**
     * What's the string used to quote SQL identifiers?
     * This returns a space " " if identifier quoting isn't supported.
     *
     * A JDBC-Compliant driver always uses a double quote character.
     *
     * @return the quoting string
     * @exception SQLException if a database-access error occurs.
     */
    public String getIdentifierQuoteString()
    throws SQLException {
        return "\"";
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public String getSQLKeywords()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public String getNumericFunctions()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public String getStringFunctions()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public String getSystemFunctions()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public String getTimeDateFunctions()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * This is the string that can be used to escape '_' or '%' in
     * the string pattern style catalog search parameters.
     *
     * <P>The '_' character represents any single character.
     * <P>The '%' character represents any sequence of zero or
     * more characters.
     *
     * @return the string used to escape wildcard characters
     * @exception SQLException if a database-access error occurs.
     */
    public String getSearchStringEscape()
    throws SQLException {
        return "\\";
    }

    /**
     * Get all the "extra" characters that can be used in unquoted
     * identifier names (those beyond a-z, A-Z, 0-9 and _).
     *
     * @return the string containing the extra characters
     * @exception SQLException if a database-access error occurs.
     */
    public String getExtraNameCharacters()
    throws SQLException {
        return "";
    }

    //--------------------------------------------------------------------
    // Functions describing which features are supported.

    /**
     * Is "ALTER TABLE" with add column supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsAlterTableWithAddColumn()
    throws SQLException {
        return true;
    }

    /**
     * Is "ALTER TABLE" with drop column supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsAlterTableWithDropColumn()
    throws SQLException {
        return true;
    }

    /**
     * Is column aliasing supported?
     *
     * <P>If so, the SQL AS clause can be used to provide names for
     * computed columns or to provide alias names for columns as
     * required.
     *
     * A JDBC-Compliant driver always returns true.
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsColumnAliasing()
    throws SQLException {
        return true;
    }

    /**
     * Are concatenations between NULL and non-NULL values NULL?
     *
     * A JDBC-Compliant driver always returns true.
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean nullPlusNonNullIsNull()
    throws SQLException {
        return true;
    }

    /**
     * Is the CONVERT function between SQL types supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsConvert()
    throws SQLException {
        return false;
    }

    /**
     * Is CONVERT between the given SQL types supported?
     *
     * @param fromType the type to convert from
     * @param toType the type to convert to
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     * @see Types
     */
    public boolean supportsConvert(int fromType, int toType)
    throws SQLException {
        return false;
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean supportsTableCorrelationNames()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean supportsDifferentTableCorrelationNames()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * Are expressions in "ORDER BY" lists supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsExpressionsInOrderBy()
    throws SQLException {
        return true;
    }

    /**
     * Can an "ORDER BY" clause use columns not in the SELECT?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsOrderByUnrelated()
    throws SQLException {
        return true;
    }

    /**
     * Is some form of "GROUP BY" clause supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsGroupBy()
    throws SQLException {
        return true;
    }

    /**
     * Can a "GROUP BY" clause use columns not in the SELECT?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsGroupByUnrelated()
    throws SQLException {
        return true;
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean supportsGroupByBeyondSelect()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * Is the escape character in "LIKE" clauses supported?
     *
     * A JDBC-Compliant driver always returns true.
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsLikeEscapeClause()
    throws SQLException {
        return true;
    }

    /**
     * Are multiple ResultSets from a single execute supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsMultipleResultSets()
    throws SQLException {
        return false;
    }

    /**
     * Can we have multiple transactions open at once (on different
     * connections)?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsMultipleTransactions()
    throws SQLException {
        return true;
    }

    /**
     * Can columns be defined as non-nullable?
     *
     * A JDBC-Compliant driver always returns true.
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsNonNullableColumns()
    throws SQLException {
        return true;
    }

    /**
     * Is the ODBC Minimum SQL grammar supported?
     *
     * All JDBC-Compliant drivers must return true.
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsMinimumSQLGrammar()
    throws SQLException {
        return true;
    }

    /**
     * Is the ODBC Core SQL grammar supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsCoreSQLGrammar()
    throws SQLException {
        return false;  /* need SQL 92 */
    }

    /**
     * Is the ODBC Extended SQL grammar supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsExtendedSQLGrammar()
    throws SQLException {
        return false;
    }

    /**
     * Is the ANSI92 entry level SQL grammar supported?
     *
     * All JDBC-Compliant drivers must return true.
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsANSI92EntryLevelSQL()
    throws SQLException {
        return supportsOuterJoins();
    }

    /**
     * Is the ANSI92 intermediate SQL grammar supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsANSI92IntermediateSQL()
    throws SQLException {
        return false;
    }

    /**
     * Is the ANSI92 full SQL grammar supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsANSI92FullSQL()
    throws SQLException {
        return false;
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean supportsIntegrityEnhancementFacility()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * Is some form of outer join supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsOuterJoins()
    throws SQLException {
        return false;
    }

    /**
     * Are full nested outer joins supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsFullOuterJoins()
    throws SQLException {
        return false;
    }

    /**
     * Is there limited support for outer joins?  (This will be true
     * if supportFullOuterJoins is true.)
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsLimitedOuterJoins()
    throws SQLException {
        return false;
    }

    /**
     * What's the database vendor's preferred term for "schema"?
     *
     * @return the vendor term
     * @exception SQLException if a database-access error occurs.
     */
    public String getSchemaTerm()
    throws SQLException {
        return "schema";
    }

    /**
     * What's the database vendor's preferred term for "procedure"?
     *
     * @return the vendor term
     * @exception SQLException if a database-access error occurs.
     */
    public String getProcedureTerm()
    throws SQLException {
        return "procedure";
    }

    /**
     * What's the database vendor's preferred term for "catalog"?
     *
     * @return the vendor term
     * @exception SQLException if a database-access error occurs.
     */
    public String getCatalogTerm()
    throws SQLException {
        return "catalog";
    }

    /**
     * Does a catalog appear at the start of a qualified table name?
     * (Otherwise it appears at the end)
     *
     * @return true if it appears at the start
     * @exception SQLException if a database-access error occurs.
     */
    public boolean isCatalogAtStart()
    throws SQLException {
        return false; /* no catalogs at all, actually. */
    }

    /**
     * What's the separator between catalog and table name?
     *
     * @return the separator string
     * @exception SQLException if a database-access error occurs.
     */
    public String getCatalogSeparator()
    throws SQLException {
        return "";  /* no catalogs */
    }

    /**
     * Can a schema name be used in a data manipulation statement?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsSchemasInDataManipulation()
    throws SQLException {
        return false;
    }

    /**
     * Can a schema name be used in a procedure call statement?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsSchemasInProcedureCalls()
    throws SQLException {
        return false;
    }

    /**
     * Can a schema name be used in a table definition statement?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsSchemasInTableDefinitions()
    throws SQLException {
        return false;
    }

    /**
     * Can a schema name be used in an index definition statement?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsSchemasInIndexDefinitions()
    throws SQLException {
        return false;
    }

    /**
     * Can a schema name be used in a privilege definition statement?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsSchemasInPrivilegeDefinitions()
    throws SQLException {
        return false;
    }

    /**
     * Can a catalog name be used in a data manipulation statement?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsCatalogsInDataManipulation()
    throws SQLException {
        return false;
    }

    /**
     * Can a catalog name be used in a procedure call statement?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsCatalogsInProcedureCalls()
    throws SQLException {
        return false;
    }

    /**
     * Can a catalog name be used in a table definition statement?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsCatalogsInTableDefinitions()
    throws SQLException {
        return false;
    }

    /**
     * Can a catalog name be used in an index definition statement?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsCatalogsInIndexDefinitions()
    throws SQLException {
        return false;
    }

    /**
     * Can a catalog name be used in a privilege definition statement?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsCatalogsInPrivilegeDefinitions()
    throws SQLException {
        return false;
    }


    /**
     * Is positioned DELETE supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsPositionedDelete()
    throws SQLException {
        return false;  /* on the TO-DO list */
    }

    /**
     * Is positioned UPDATE supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsPositionedUpdate()
    throws SQLException {
        return false;
    }

    /**
     * Is SELECT for UPDATE supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsSelectForUpdate()
    throws SQLException {
        return true;
    }

    /**
     * Are stored procedure calls using the stored procedure escape
     * syntax supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsStoredProcedures()
    throws SQLException {
        return false;  /* TODO: should implement this */
    }

    /**
     * Are subqueries in comparison expressions supported?
     *
     * A JDBC-Compliant driver always returns true.
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsSubqueriesInComparisons()
    throws SQLException {
        return true;
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean supportsSubqueriesInExists()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * Are subqueries in 'in' statements supported?
     *
     * A JDBC-Compliant driver always returns true.
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsSubqueriesInIns()
    throws SQLException {
        return true;
    }

    /**
     * Are subqueries in quantified expressions supported?
     *
     * A JDBC-Compliant driver always returns true.
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsSubqueriesInQuantifieds()
    throws SQLException {
        return true;
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean supportsCorrelatedSubqueries()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * Is SQL UNION supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsUnion()
    throws SQLException {
        return true;
    }

    /**
     * Is SQL UNION ALL supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsUnionAll()
    throws SQLException {
        return true;
    }

    /**
     * Can cursors remain open across commits?
     *
     * @return true if cursors always remain open; false if they might not remain open
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsOpenCursorsAcrossCommit()
    throws SQLException {
        return false;
    }

    /**
     * Can cursors remain open across rollbacks?
     *
     * @return true if cursors always remain open; false if they might not remain open
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsOpenCursorsAcrossRollback()
    throws SQLException {
        return false;
    }

    /**
     * Can statements remain open across commits?
     *
     * @return true if statements always remain open; false if they might not remain open
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsOpenStatementsAcrossCommit()
    throws SQLException {
        return true;
    }

    /**
     * Can statements remain open across rollbacks?
     *
     * @return true if statements always remain open; false if they might not remain open
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsOpenStatementsAcrossRollback()
    throws SQLException {
        return true;
    }



    //----------------------------------------------------------------------
    // The following group of methods exposes various limitations
    // based on the target database with the current driver.
    // Unless otherwise specified, a result of zero means there is no
    // limit, or the limit is not known.

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public int getMaxBinaryLiteralLength()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public int getMaxCharLiteralLength()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * What's the limit on column name length?
     *
     * @return max literal length
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxColumnNameLength()
    throws SQLException {
        return MAX_IDENTIFIER_LENGTH;
    }

    /**
     * What's the maximum number of columns in a "GROUP BY" clause?
     *
     * @return max number of columns
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxColumnsInGroupBy()
    throws SQLException {
        return getMaxColumnsInTable();
    }

    /**
     * What's the maximum number of columns allowed in an index?
     *
     * @return max columns
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxColumnsInIndex()
    throws SQLException {
        return getMaxColumnsInTable();
    }

    /**
     * What's the maximum number of columns in an "ORDER BY" clause?
     *
     * @return max columns
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxColumnsInOrderBy()
    throws SQLException {
        return getMaxColumnsInTable();
    }

    /**
     * What's the maximum number of columns in a "SELECT" list?
     *
     * @return max columns
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxColumnsInSelect()
    throws SQLException {
        return getMaxColumnsInTable();
    }

    /**
     * What's the maximum number of columns in a table?
     *
     * @return max columns
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxColumnsInTable()
    throws SQLException {
        return 1600;
    }

    /**
     * How many active connections can we have at a time to this database?
     *
     * @return max connections
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxConnections()
    throws SQLException {
        return 1024;  /* absolute maximum.  configured at startup.
                       * default is 32 it seems */
    }

    /**
     * What's the maximum cursor name length?
     *
     * @return max cursor name length in bytes
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxCursorNameLength()
    throws SQLException {
        return MAX_IDENTIFIER_LENGTH;
    }

    /**
     * What's the maximum length of an index (in bytes)?
     *
     * @return max index length in bytes
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxIndexLength()
    throws SQLException {
        return MAX_ROW_SIZE;
    }

    /**
     * What's the maximum length allowed for a schema name?
     *
     * @return max name length in bytes
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxSchemaNameLength()
    throws SQLException {
        return 0;
    }

    /**
     * What's the maximum length of a procedure name?
     *
     * @return max name length in bytes
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxProcedureNameLength()
    throws SQLException {
        return MAX_IDENTIFIER_LENGTH;
    }

    /**
     * What's the maximum length of a catalog name?
     *
     * @return max name length in bytes
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxCatalogNameLength()
    throws SQLException {
        return 0;
    }

    /**
     * What's the maximum length of a single row?
     *
     * @return max row size in bytes
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxRowSize()
    throws SQLException {
        return MAX_ROW_SIZE;
    }

    /**
     * Did getMaxRowSize() include LONGVARCHAR and LONGVARBINARY
     * blobs?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean doesMaxRowSizeIncludeBlobs()
    throws SQLException {
        return false;
    }

    /**
     * What's the maximum length of a SQL statement?
     *
     * @return max length in bytes
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxStatementLength()
    throws SQLException {
        return MAX_STATEMENT_LENGTH;
    }

    /**
     * How many active statements can we have open at one time to this
     * database?
     *
     * @return the maximum
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxStatements()
    throws SQLException {
        return 0;
    }

    /**
     * What's the maximum length of a table name?
     *
     * @return max name length in bytes
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxTableNameLength()
    throws SQLException {
        return MAX_IDENTIFIER_LENGTH;
    }

    /**
     * What's the maximum number of tables in a SELECT?
     *
     * @return the maximum
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxTablesInSelect()
    throws SQLException {
        return 0;  /* unlimited (except for query length) */
    }

    /**
     * What's the maximum length of a user name?
     *
     * @return max name length  in bytes
     * @exception SQLException if a database-access error occurs.
     */
    public int getMaxUserNameLength()
    throws SQLException {
        return MAX_USERNAME_LENGTH;
    }

    //----------------------------------------------------------------------

    /**
     * What's the database's default transaction isolation level?  The
     * values are defined in java.sql.Connection.
     *
     * @return the default isolation level
     * @exception SQLException if a database-access error occurs.
     * @see Connection
     */
    public int getDefaultTransactionIsolation()
    throws SQLException {
        return Connection.TRANSACTION_READ_COMMITTED; /* according to manual */
    }

    /**
     * Are transactions supported? If not, commit is a noop and the
     * isolation level is TRANSACTION_NONE.
     *
     * @return true if transactions are supported
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsTransactions()
    throws SQLException {
        return true;
    }

    /**
     * Does the database support the given transaction isolation level?
     *
     * @param level the values are defined in java.sql.Connection
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     * @see Connection
     */
    public boolean supportsTransactionIsolationLevel(int level)
    throws SQLException {
        return (level == Connection.TRANSACTION_READ_COMMITTED
                || level == Connection.TRANSACTION_SERIALIZABLE);
    }

    /**
     * Are both data definition and data manipulation statements
     * within a transaction supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsDataDefinitionAndDataManipulationTransactions()
    throws SQLException {
        return true;  /* was able to rollback a table creation. */
    }

    /**
     * Are only data manipulation statements within a transaction
     * supported?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean supportsDataManipulationTransactionsOnly()
    throws SQLException {
        return !supportsDataDefinitionAndDataManipulationTransactions();
    }

    /**
     * Does a data definition statement within a transaction force the
     * transaction to commit?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean dataDefinitionCausesTransactionCommit()
    throws SQLException {
        return false;
    }

    /**
     * Is a data definition statement within a transaction ignored?
     *
     * @return true if so
     * @exception SQLException if a database-access error occurs.
     */
    public boolean dataDefinitionIgnoredInTransactions()
    throws SQLException {
        return false;
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getProcedures(String catalog,
                                            String schemaPattern,
                                            String procedureNamePattern)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * PROCEDURE_TYPE - May return a result.
     */
    public static final int procedureResultUnknown = 0;
    /**
     * PROCEDURE_TYPE - Does not return a result.
     */
    public static final int procedureNoResult = 1;
    /**
     * PROCEDURE_TYPE - Returns a result.
     */
    public static final int procedureReturnsResult = 2;

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getProcedureColumns(String catalog,
                                         String schemaPattern,
                                         String procedureNamePattern,
                                         String columnNamePattern)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * COLUMN_TYPE - nobody knows.
     */
    public static final int procedureColumnUnknown = 0;

    /**
     * COLUMN_TYPE - IN parameter.
     */
    public static final int procedureColumnIn = 1;

    /**
     * COLUMN_TYPE - INOUT parameter.
     */
    public static final int procedureColumnInOut = 2;

    /**
     * COLUMN_TYPE - OUT parameter.
     */
    public static final int procedureColumnOut = 4;
    /**
     * COLUMN_TYPE - procedure return value.
     */
    public static final int procedureColumnReturn = 5;

    /**
     * COLUMN_TYPE - result column in ResultSet.
     */
    public static final int procedureColumnResult = 3;

    /**
     * TYPE NULLABLE - does not allow NULL values.
     */
    public static final int procedureNoNulls = 0;

    /**
     * TYPE NULLABLE - allows NULL values.
     */
    public static final int procedureNullable = 1;

    /**
     * TYPE NULLABLE - nullability unknown.
     */
    public static final int procedureNullableUnknown = 2;

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getTables(String catalog,
                               String schemaPattern,
                               String tableNamePattern,
                               String types[])
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getSchemas()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getCatalogs()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getTableTypes()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getColumns(String catalog,
                                String schemaPattern,
                                String tableNamePattern,
                                String columnNamePattern)

    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }
    /**
     * COLUMN NULLABLE - might not allow NULL values.
     */
    public static final int columnNoNulls = 0;

    /**
     * COLUMN NULLABLE - definitely allows NULL values.
     */
    public static final int columnNullable = 1;

    /**
     * COLUMN NULLABLE - nullability unknown.
     */
    public static final int columnNullableUnknown = 2;

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getColumnPrivileges(String catalog,
                                                  String schema,
                                                  String table,
                                                  String columnNamePattern)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getTablePrivileges(String catalog,
                                                 String schemaPattern,
                                                 String tableNamePattern)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getBestRowIdentifier(String catalog,
                                                   String schema,
                                                   String table, int scope,
                                                   boolean nullable)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * BEST ROW SCOPE - very temporary, while using row.
     */
    public static final int bestRowTemporary   = 0;

    /**
     * BEST ROW SCOPE - valid for remainder of current transaction.
     */
    public static final int bestRowTransaction = 1;

    /**
     * BEST ROW SCOPE - valid for remainder of current session.
     */
    public static final int bestRowSession     = 2;

    /**
     * BEST ROW PSEUDO_COLUMN - may or may not be pseudo column.
     */
    public static final int bestRowUnknown      = 0;

    /**
     * BEST ROW PSEUDO_COLUMN - is NOT a pseudo column.
     */
    public static final int bestRowNotPseudo    = 1;

    /**
     * BEST ROW PSEUDO_COLUMN - is a pseudo column.
     */
    public static final int bestRowPseudo       = 2;

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getVersionColumns(String catalog, String schema,
                                                String table)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * VERSION COLUMNS PSEUDO_COLUMN - may or may not be pseudo column.
     */
    public static final int versionColumnUnknown        = 0;

    /**
     *  VERSION COLUMNS PSEUDO_COLUMN - is NOT a pseudo column.
     */
    public static final int versionColumnNotPseudo      = 1;

    /**
     *  VERSION COLUMNS PSEUDO_COLUMN - is a pseudo column.
     */
    public static final int versionColumnPseudo = 2;

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getPrimaryKeys(String catalog, String schema,
                                             String table)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getImportedKeys(String catalog, String schema,
                                     String table)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * IMPORT KEY UPDATE_RULE and DELETE_RULE - for update, change
     * imported key to agree with primary key update; for delete,
     * delete rows that import a deleted key.
     */
    public static final int importedKeyCascade  = 0;

    /**
     * IMPORT KEY UPDATE_RULE and DELETE_RULE - do not allow update or
     * delete of primary key if it has been imported.
     */
    public static final int importedKeyRestrict = 1;

    /**
     * IMPORT KEY UPDATE_RULE and DELETE_RULE - change imported key to
     * NULL if its primary key has been updated or deleted.
     */
    public static final int importedKeySetNull  = 2;

    /**
     * IMPORT KEY UPDATE_RULE and DELETE_RULE - do not allow update or
     * delete of primary key if it has been imported.
     */
    public static final int importedKeyNoAction = 3;

    /**
     * IMPORT KEY UPDATE_RULE and DELETE_RULE - change imported key to
     * default values if its primary key has been updated or deleted.
     */
    public static final int importedKeySetDefault  = 4;

    /**
     * IMPORT KEY DEFERRABILITY - see SQL92 for definition
     */
    public static final int importedKeyInitiallyDeferred  = 5;

    /**
     * IMPORT KEY DEFERRABILITY - see SQL92 for definition
     */
    public static final int importedKeyInitiallyImmediate  = 6;

    /**
     * IMPORT KEY DEFERRABILITY - see SQL92 for definition
     */
    public static final int importedKeyNotDeferrable  = 7;

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getExportedKeys(String catalog, String schema,
                                              String table)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getCrossReference(String primaryCatalog,
                                                String primarySchema,
                                                String primaryTable,
                                                String foreignCatalog,
                                                String foreignSchema,
                                                String foreignTable)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getTypeInfo()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * TYPE NULLABLE - does not allow NULL values.
     */
    public static final int typeNoNulls = 0;

    /**
     * TYPE NULLABLE - allows NULL values.
     */
    public static final int typeNullable = 1;

    /**
     * TYPE NULLABLE - nullability unknown.
     */
    public static final int typeNullableUnknown = 2;

    /**
     * TYPE INFO SEARCHABLE - No support.
     */
    public static final int typePredNone = 0;

    /**
     * TYPE INFO SEARCHABLE - Only supported with WHERE .. LIKE.
     */
    public static final int typePredChar = 1;

    /**
     * TYPE INFO SEARCHABLE -  Supported except for WHERE .. LIKE.
     */
    public static final int typePredBasic = 2;

    /**
     * TYPE INFO SEARCHABLE - Supported for all WHERE ...
     */
    public static final int typeSearchable  = 3;

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getIndexInfo(String catalog, String schema,
                                           String table,
                                           boolean unique, boolean approximate)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * INDEX INFO TYPE - this identifies table statistics that are
     * returned in conjuction with a table's index descriptions.
     */
    public static final short tableIndexStatistic = 0;

    /**
     * INDEX INFO TYPE - this identifies a clustered index.
     */
    public static final short tableIndexClustered = 1;

    /**
     * INDEX INFO TYPE - this identifies a hashed index.
     */
    public static final short tableIndexHashed    = 2;

    /**
     * INDEX INFO TYPE - this identifies some other form of index.
     */
    public static final short tableIndexOther     = 3;

    /* -- JDBC 2.0 ---------------------------------------------------- */

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean supportsResultSetType(int type)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean supportsResultSetConcurrency(int type, int concurrency)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean ownUpdatesAreVisible(int type)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean ownDeletesAreVisible(int type)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean ownInsertsAreVisible(int type)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean othersUpdatesAreVisible(int type)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean othersDeletesAreVisible(int type)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean othersInsertsAreVisible(int type)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean updatesAreDetected(int type)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean deletesAreDetected(int type)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean insertsAreDetected(int type)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean supportsBatchUpdates()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.ResultSet getUDTs(String catalog, String schemaPattern,
                                      String typeNamePattern, int[] types)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.Connection getConnection()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }
}
