/*------------------------------------------------------------------------
 |  FILE            Driver.java
 |  PART OF         Sverre's Simple JDBC Driver for PostgreSQL
 |  WRITTEN BY      Sverre H. Huseby <shh@thathost.com>
 +----------------------------------------------------------------------*/

package no.shhsoft.postgresql;

import java.util.*;
import java.sql.*;

/**
 * The Java SQL framework allows for multiple database drivers.
 * <P>
 * Each driver should supply a class that implements the Driver
 * interface.
 * <P>
 * The DriverManager will try to load as many drivers as it can find
 * and then for any given connection request, it will ask each driver
 * in turn to try to connect to the target URL.
 * <P>
 * It is strongly recommended that each Driver class should be small
 * and standalone so that the Driver class can be loaded and queried
 * without bringing in vast quantities of supporting code.
 * <P>
 * When a Driver class is loaded, it should create an instance of
 * itself and register it with the DriverManager. This means that a
 * user can load and register a driver by doing
 * Class.forName("foo.bah.Driver").
 *
 * @see      DriverManager
 * @see      Connection
 * @author   Sverre H. Huseby
 *           &lt;<A HREF="mailto:shh@thathost.com">shh@thathost.com</A>&gt;
 * @version  $Id: Driver.java,v 1.4 2002/02/05 22:54:25 sverrehu Exp $
 */
public final class Driver
implements java.sql.Driver {
/*-----------------------------------------------------------------------+
|  PRIVATE PART                                                          |
+-----------------------------------------------------------------------*/
    private static final String URL_START = "jdbc:SHHPostgreSQL:";

    static {
        try {
            java.sql.DriverManager.registerDriver(new Driver());
        } catch (SQLException e) {
            System.err.println("unable to register JDBC driver: "
                               + e.getMessage());
        }
    }

    private Properties urlToProperties(String url, Properties deflt)
    throws SQLException {
        String     rest, userHostPart, hostPart;
        int        idx, idx2, idx3;
        Properties ret;

        if (!url.startsWith(URL_START))
            return null;
        rest = url.substring(URL_START.length());
        ret = new Properties(deflt);

        if (rest.startsWith("//")) {
            /* includes [user[=password]@]host */
            rest = rest.substring(2);
            if ((idx = rest.indexOf('/')) < 0)
                throw new SQLException("Invaild URL: Missing database name");
            userHostPart = rest.substring(0, idx);
            rest = rest.substring(idx + 1);

            idx = userHostPart.indexOf('=');
            idx2 = userHostPart.indexOf('@');
            if (idx >= 0 && idx < idx2) {
                /* includes user=password@ */
                ret.put("user", userHostPart.substring(0, idx));
                ret.put("password", userHostPart.substring(idx + 1, idx2));
                hostPart = userHostPart.substring(idx2 + 1);
            } else if (idx2 >= 0) {
                /* include user@ */
                ret.put("user", userHostPart.substring(idx2));
                hostPart = userHostPart.substring(idx2 + 1);
            } else
                hostPart = userHostPart;

            idx = hostPart.indexOf(':');
            if (idx >= 0) {
                ret.put("host", hostPart.substring(0, idx));
                ret.put("port", hostPart.substring(idx + 1));
            } else
                ret.put("host", hostPart);
        }

        ret.put("database", rest);

        return ret;
    }

/*-----------------------------------------------------------------------+
|  PACKAGE LOCAL PART                                                    |
+-----------------------------------------------------------------------*/
    static final int DRIVER_MAJOR_VERSION = 0;
    static final int DRIVER_MINOR_VERSION = 1;

/*-----------------------------------------------------------------------+
|  PUBLIC INTERFACE                                                      |
+-----------------------------------------------------------------------*/

    /* java.sql.Driver **************************************************/
    /**
     * Attempts to make a database connection to the given URL.  The
     * driver should return "null" if it realizes it is the wrong kind
     * of driver to connect to the given URL.  This will be common, as
     * when the JDBC driver manager is asked to connect to a given URL
     * it passes the URL to each loaded driver in turn.
     *
     * <P>The driver should raise a SQLException if it is the right
     * driver to connect to the given URL, but has trouble connecting
     * to the database.
     *
     * <P>The java.util.Properties argument can be used to passed
     * arbitrary string tag/value pairs as connection arguments.
     * Normally at least "user" and "password" properties should be
     * included in the Properties.
     *
     * @param      url     The URL of the database to connect to
     * @param      info    a list of arbitrary string tag/value pairs as
     *                     connection arguments; normally at least
     *                     a "user" and "password" property should be
     *                     included
     * @return     a Connection to the URL
     * @exception  SQLException if a database-access error occurs.
     */
    public java.sql.Connection connect(String url, java.util.Properties info)
    throws SQLException {
        Properties props;
        String     s;
        int        port = -1;

        if ((props = urlToProperties(url, info)) == null)
            return null;
        if ((s = props.getProperty("port")) != null) {
            try {
                port = Integer.parseInt(s);
            } catch (NumberFormatException e) {
                throw new SQLException("Invalid port number in database URL");
            }
        }
        return new Connection(url,
                              props.getProperty("host"), port,
                              props.getProperty("database"),
                              props.getProperty("user"),
                              props.getProperty("password"));
    }

    /**
     * Returns true if the driver thinks that it can open a connection
     * to the given URL.  Typically drivers will return true if they
     * understand the subprotocol specified in the URL and false if
     * they don't.
     *
     * @param      url     The URL of the database.
     * @return     True if this driver can connect to the given URL.
     * @exception  SQLException if a database-access error occurs.
     */
    public boolean acceptsURL(String url)
    throws SQLException {
        try {
            return (urlToProperties(url, null) != null);
        } catch (SQLException e) {
        }
        return false;
    }

    /**
     * Gets information about the possible properties for this driver.
     * <p>
     * The getPropertyInfo method is intended to allow a generic
     * GUI tool to discover what properties it should prompt a human
     * for in order to get enough information to connect to a
     * database.  Note that depending on the values the human has
     * supplied so far, additional values may become necessary, so it
     * may be necessary to iterate though several calls to
     * getPropertyInfo.
     *
     * @param      url     The URL of the database to connect to.
     * @param      info    A proposed list of tag/value pairs that will
     *                     be sent on connect open.
     * @return     An array of DriverPropertyInfo objects describing
     *             possible properties.  This array may be an empty array
     *             if no properties are required.
     * @exception  SQLException if a database-access error occurs.
     */
    public DriverPropertyInfo[] getPropertyInfo(String url,
                                                java.util.Properties info)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * Gets the driver's major version number. Initially this should be 1.
     */
    public int getMajorVersion() {
        return DRIVER_MAJOR_VERSION;
    }

    /**
     * Gets the driver's minor version number. Initially this should be 0.
     */
    public int getMinorVersion() {
        return DRIVER_MINOR_VERSION;
    }

    /**
     * Reports whether the Driver is a genuine JDBC COMPLIANT (tm)
     * driver.  A driver may only report "true" here if it passes the
     * JDBC compliance tests, otherwise it is required to return
     * false.
     *
     * JDBC compliance requires full support for the JDBC API and full
     * support for SQL 92 Entry Level.  It is expected that JDBC
     * compliant drivers will be available for all the major
     * commercial databases.
     *
     * This method is not intended to encourage the development of
     * non-JDBC compliant drivers, but is a recognition of the fact
     * that some vendors are interested in using the JDBC API and
     * framework for lightweight databases that do not support full
     * database functionality, or for special databases such as
     * document information retrieval where a SQL implementation may
     * not be feasible.
     */
    public boolean jdbcCompliant() {
        return false;  /* PostgreSQL is not SQL92 compliant. */
    }
}
