/*------------------------------------------------------------------------
 |  FILE            PostgreSQLCommunicator.java
 |  PART OF         Sverre's Simple JDBC Driver for PostgreSQL
 |  WRITTEN BY      Sverre H. Huseby <shh@thathost.com>
 +----------------------------------------------------------------------*/

package no.shhsoft.postgresql;

import java.util.*;
import java.net.*;
import java.io.*;
import java.sql.*;

/**
 * The communication protocol implemented in this class is documented
 * in the PostgreSQL Programmer's Guide, in the section named
 * "Frontend/Backend Protocol".
 *
 * @author   Sverre H. Huseby
 *           &lt;<A HREF="mailto:shh@thathost.com">shh@thathost.com</A>&gt;
 * @version  $Id: PostgreSQLCommunicator.java,v 1.10 2002/02/05 23:19:40 sverrehu Exp $
 */
final class PostgreSQLCommunicator {
/*-----------------------------------------------------------------------+
|  PRIVATE PART                                                          |
+-----------------------------------------------------------------------*/
    private static final int DEFAULT_PORT = 5432;
    private static final int OUTPUT_BUFFER_SIZE = 16384;
    private static final int INPUT_BUFFER_SIZE  = 16384;

    private String               host;
    private int                  port;
    private Socket               sock;
    private boolean              socketConnected = false;
    private boolean              databaseConnected = false;
    private BufferedOutputStream out;
    private BufferedInputStream  in;

    private int backendProcessId = -1;
    private int backendSecretKey = -1;

    private boolean lastCodeWasReadyForQuery = false;

    /* the following is used within repeated calls to
     * getAndProcessCommand. */
    private boolean gotARow;

    private void closeSocket()
    throws IOException {
        if (!socketConnected)
            return;
        databaseConnected = false;
        socketConnected = false;
        sock.close();
    }

    private void sendInt32(int i)
    throws IOException {
        out.write((byte) ((i >> 24) & 0xFF));
        out.write((byte) ((i >> 16) & 0xFF));
        out.write((byte) ((i >>  8) & 0xFF));
        out.write((byte) ( i        & 0xFF));
    }

    private int receiveInt32()
    throws IOException {
        int ret;

        ret = in.read();
        ret = (ret << 8) | in.read();
        ret = (ret << 8) | in.read();
        ret = (ret << 8) | in.read();
        return ret;
    }

    private void sendInt16(int i)
    throws IOException {
        out.write((byte) ((i >>  8) & 0xFF));
        out.write((byte) ( i        & 0xFF));
    }

    private int receiveInt16()
    throws IOException {
        int ret;

        ret = in.read();
        ret = (ret << 8) | in.read();
        return ret;
    }

    private void sendLimString(int n, String s)
    throws IOException {
        byte[] buff;
        int    slen;

        buff = new byte[n];
        slen = s.length();
        if (slen > n)
            slen = n;
        else
            buff[n - 1] = 0;
        s.getBytes(0, slen, buff, 0);
        out.write(buff);
    }

    private void sendString(String s)
    throws IOException {
        out.write(s.getBytes());
        out.write('\0');
    }

    private String receiveString()
    throws IOException {
        StringBuffer sb;
        int          c;

        sb = new StringBuffer();
        while ((c = in.read()) > 0)
            sb.append((char) c);
        return sb.toString();
    }

    private void receiveByteN(int n, byte[] buff)
    throws IOException {
        int nread, totread = 0;

        while (totread < n) {
            nread = in.read(buff, totread, n - totread);
            totread += nread;
        }
    }

    private void skip(int n)
    throws IOException {
        long nskipped, totskipped = 0L;

        while (totskipped < n) {
            nskipped = in.skip(n - totskipped);
            totskipped += nskipped;
        }
    }

    private int getCommand()
    throws IOException, SQLException {
        int    code;
        String msg;

        for (;;) {
            code = in.read();
            if (code == 'E') {
                /* ErrorResponse */
                msg = receiveString();
                throw new SQLException(msg);
            } else if (code == 'N') {
                /* NoticeResponse */
                msg = receiveString();
                System.err.println(msg);
            } else
                break;
        }
        return code;
    }

    private static final int QUERY_NOT_DONE = -1;
    private static final int QUERY_DONE     = -2;

    /* return value >= 0 indicates an update count.  otherwise returns
     * QUERY_NOT_DONE or QUERY_DONE. */
    private int getAndProcessCommand(ResultSet rs)
    throws IOException, SQLException {
        int q, code;
        int ret = QUERY_NOT_DONE;

        code = getCommand();
        lastCodeWasReadyForQuery = false;
        switch (code) {
          case 'B': /* BinaryRow */
            gotARow = true;
            throw new NotImplementedException("Binary data from the backend.");
            /* TODO */
            // break;
          case 'C': /* CompletedResponse */
            /* an SQL command (a query may contain several commands)
             * completed normally. */
            String s;

            s = receiveString(); /* tag */
            try {
                ret = Integer.parseInt(s.substring(s.lastIndexOf(' ')));
            } catch (Exception e) {
                /* no update count found.  probably not an update reply,
                 * then.  just ignore it. */
            }
            if (rs != null && !gotARow)
                rs.fetchWithoutResult = true;
            break;
          case 'D': /* AsciiRow */
            gotARow = true;
            if (rs == null)
                throw new InternalErrorException("Unexpected AsciiRow");
            receiveByteN(rs.notNullMapBytes, rs.notNullMap);
            int maskBit = 0x80;
            int mapByte = 0;
            int size;
            byte[][] row;
            row = rs.field[rs.numRows];
            for (q = 0; q < rs.numFields; q++) {
                if ((rs.notNullMap[mapByte] & maskBit) != 0) {
                    /* not null field */
                    size = receiveInt32() - 4;
                    row[q] = new byte[size];
                    receiveByteN(size, row[q]);
                } else {
                    /* null field */
                    row[q] = null;
                }
                if ((maskBit >>= 1) == 0) {
                    maskBit = 0x80;
                    ++mapByte;
                }
            }
            ++rs.numRows;
            break;
          case 'G': /* CopyInResponse */
            throw new NotImplementedException("CopyInResponse");
            // break;
          case 'H': /* CopyOutResponse */
            throw new NotImplementedException("CopyOutResponse");
            /* TODO: a CopyDataRows message will follow. */
            // break;
          case 'I': /* EmptyQuery */
            receiveString(); /* unused */
            break;
          case 'P': /* CursorResponse */
            receiveString(); /* cursor name */
            /* a CompletedResponse or RowDescription will follow. */
            break;
          case 'T': /* RowDescription */
            if (rs == null)
                throw new InternalErrorException("Unexpected RowDescription");
            int numFields;
            numFields = receiveInt16();
            if (rs.numFields < 0) {
                rs.numFields = numFields;
                rs.fieldName = new String[numFields];
                rs.fieldNameToIdx = new Hashtable();
                rs.fieldTypeOID = new int[numFields];
                rs.fieldTypeSize = new int[numFields];
                rs.fieldTypeModifier = new int[numFields];

                for (q = 0; q < numFields; q++) {
                    rs.fieldName[q] = receiveString();
                    rs.fieldNameToIdx.put(rs.fieldName[q], new Integer(q + 1));
                    rs.fieldTypeOID[q] = receiveInt32();
                    rs.fieldTypeSize[q] = receiveInt16();
                    rs.fieldTypeModifier[q] = receiveInt32();
                }
                if (numFields > 0) {
                    rs.notNullMapBytes = ((numFields - 1) >> 3) + 1;
                    rs.notNullMap = new byte[rs.notNullMapBytes];
                } else {
                    rs.notNullMapBytes = 0;
                }

                for (q = rs.fetchSize - 1; q >= 0; q--)
                    rs.field[q] = new byte[numFields][];
            } else {
                /* know it, so we just skip it. */
                for (q = 0; q < numFields; q++) {
                    receiveString();
                    skip(10);
                }
                break;
            }
            /* AsciiRow or BinaryRow will follow. */
            break;
          case 'Z': /* ReadyForQuery */
            /* received when everything is finished, regardless of
             * errors. */
            lastCodeWasReadyForQuery = true;
            ret = QUERY_DONE;
            break;
          default:
            throw new NeedUpdateException("Unknown response from database: `"
                                          + (char) code + "'");
        }
        return ret;
    }

    /* return value true indicates that the command was a ReadyForQuery. */
    private boolean getAndSkipCommand(ResultSet rs)
    throws IOException, SQLException {
        int     q, code;
        boolean ret = false;
        int     maskBit = 0x80, mapByte = 0, size;

        code = getCommand();
        lastCodeWasReadyForQuery = false;
        switch (code) {
          case 'B': /* BinaryRow */
            if (rs == null)
                throw new InternalErrorException("Unexpected AsciiRow");
            receiveByteN(rs.notNullMapBytes, rs.notNullMap);
            for (q = 0; q < rs.numFields; q++) {
                if ((rs.notNullMap[mapByte] & maskBit) != 0) {
                    /* not null field */
                    size = receiveInt32() - 4;
                    skip(size);
                }
                if ((maskBit >>= 1) == 0) {
                    maskBit = 0x80;
                    ++mapByte;
                }
            }
            break;
          case 'C': /* CompletedResponse */
            receiveString(); /* tag */
            break;
          case 'D': /* AsciiRow */
            if (rs == null)
                throw new InternalErrorException("Unexpected AsciiRow");
            receiveByteN(rs.notNullMapBytes, rs.notNullMap);
            for (q = 0; q < rs.numFields; q++) {
                if ((rs.notNullMap[mapByte] & maskBit) != 0) {
                    /* not null field */
                    size = receiveInt32() - 4;
                    skip(size);
                }
                if ((maskBit >>= 1) == 0) {
                    maskBit = 0x80;
                    ++mapByte;
                }
            }
            break;
          case 'G': /* CopyInResponse */
            throw new NotImplementedException("CopyInResponse");
            // break;
          case 'H': /* CopyOutResponse */
            throw new NotImplementedException("CopyOutResponse");
            /* TODO: a CopyDataRows message will follow. */
            // break;
          case 'I': /* EmptyQuery */
            receiveString(); /* unused */
            break;
          case 'P': /* CursorResponse */
            receiveString(); /* cursor name */
            /* a CompletedResponse or RowDescription will follow. */
            break;
          case 'T': /* RowDescription */
            if (rs == null)
                throw new InternalErrorException("Unexpected RowDescription");
            int numFields;
            numFields = receiveInt16();
            if (rs.numFields < 0) {
                rs.numFields = numFields;
                rs.fieldName = new String[numFields];
                rs.fieldNameToIdx = new Hashtable();
                rs.fieldTypeOID = new int[numFields];
                rs.fieldTypeSize = new int[numFields];
                rs.fieldTypeModifier = new int[numFields];

                for (q = 0; q < numFields; q++) {
                    rs.fieldName[q] = receiveString();
                    rs.fieldNameToIdx.put(rs.fieldName[q], new Integer(q + 1));
                    rs.fieldTypeOID[q] = receiveInt32();
                    rs.fieldTypeSize[q] = receiveInt16();
                    rs.fieldTypeModifier[q] = receiveInt32();
                }
                if (numFields > 0) {
                    rs.notNullMapBytes = ((numFields - 1) >> 3) + 1;
                    rs.notNullMap = new byte[rs.notNullMapBytes];
                } else {
                    rs.notNullMapBytes = 0;
                }

                for (q = rs.fetchSize - 1; q >= 0; q--)
                    rs.field[q] = new byte[numFields][];
            } else {
                /* know it, so we just skip it. */
                for (q = 0; q < numFields; q++) {
                    receiveString();
                    skip(10);
                }
                break;
            }
            /* AsciiRow or BinaryRow will follow. */
            break;
          case 'Z': /* ReadyForQuery */
            /* received when everything is finished, regardless of
             * errors. */
            lastCodeWasReadyForQuery = true;
            ret = true;
            break;
          default:
            throw new NeedUpdateException("Unknown response from database: `"
                                          + (char) code + "'");
        }
        return ret;
    }

    private void startup(String database, String username, String password)
    throws IOException, SQLException {
        int     code, subCode;
        boolean authOK = false;
        boolean backendReady = false;
        String  msg;

        /* talk to the postmaster to log in */
        sendInt32(296); /* length */
        sendInt32((2 << 16) | 0);  /* protocol version */
        sendLimString(64, database);
        sendLimString(32, username);
        sendLimString(64, "");
        sendLimString(64, "");
        sendLimString(64, "");
        out.flush();

        do {
            try {
                code = getCommand();
            } catch (IOException e) {
                closeSocket();
                throw e;
            }
            switch (code) {
              case 'R': /* Authentication... */
                subCode = receiveInt32();
                switch (subCode) {
                  case 0: /* AuthenticationOK */
                    authOK = true;
                    break;
                  case 1: /* AuthenticationKerberosV4 */
                  case 2: /* AuthenticationKerberosV5 */
                    closeSocket();
                    throw new NeedUpdateException(
                               "Unsupported authentication method required.");
                  case 3: /* AuthenticationUnencryptedPassword */
                    sendInt32(4 + password.length() + 1);
                    sendString(password);
                    out.flush();
                    break;
                  case 4: /* AuthenticationEncryptedPassword */
                    in.read(); /* skip salt byte one */
                    in.read(); /* skip salt byte two */
                    closeSocket();
                    throw new NeedUpdateException(
                               "Unsupported authentication method required.");
                  case 5: /* AuthenticationMD5Password */
                    String md5pass;
                    byte[] salt;

                    salt = new byte[4];
                    salt[0] = (byte) in.read();
                    salt[1] = (byte) in.read();
                    salt[2] = (byte) in.read();
                    salt[3] = (byte) in.read();
                    md5pass = MD5PostgreSQL.encryptMD5(username, password,
                                                       salt);
                    sendInt32(4 + md5pass.length() + 1);
                    sendString(md5pass);
                    out.flush();
                    break;
                  default:
                    closeSocket();
                    throw new NeedUpdateException(
                               "Unknown authentication method required.");
                }
                break;
              default:
                closeSocket();
                throw new NeedUpdateException(
                               "Unknown response from backend.");
            }
        } while (!authOK);

        /* wait for backend to wake up */
        do {
            try {
                code = getCommand();
                lastCodeWasReadyForQuery = false;
            } catch (IOException e) {
                closeSocket();
                throw e;
            }
            switch (code) {
              case 'K': /* BackendKeyData */
                backendProcessId = receiveInt32(); /* backend process ID */
                backendSecretKey = receiveInt32(); /* backend secret key */
                break;
              case 'Z': /* ReadyForQuery */
                lastCodeWasReadyForQuery = true;
                backendReady = true;
                break;
              default:
                closeSocket();
                throw new NeedUpdateException(
                               "Unknown response from backend.");
            }
        } while (!backendReady);

        databaseConnected = true;
    }

    private void shutdown()
    throws IOException {
        if (!databaseConnected)
            return;
        databaseConnected = false;
        out.write('X'); /* Terminate */
        out.flush();
    }

/*-----------------------------------------------------------------------+
|  PACKAGE LOCAL PART                                                    |
+-----------------------------------------------------------------------*/
    PostgreSQLCommunicator(String host, int port, String database,
                           String username, String password)
    throws SQLException {
        if (port < 0)
            port = DEFAULT_PORT;
        if (host == null || host.length() == 0)
            host = "localhost";
        if (database == null)
            database = "";
        if (username == null)
            username = "";

        this.host = host;
        this.port = port;

        try {
            sock = new Socket(host, port);
            /* disable Nagle's algorithm (turn off time-based
             * buffering of output). */
            sock.setTcpNoDelay(true);
            socketConnected = true;

            out = new BufferedOutputStream(sock.getOutputStream(),
                                           OUTPUT_BUFFER_SIZE);
            in = new BufferedInputStream(sock.getInputStream(),
                                         INPUT_BUFFER_SIZE);

            startup(database, username, password);
            query("SET DATESTYLE TO 'ISO'");
        } catch (IOException e) {
            throw new SQLException("Unable to open connection: "
                                   + e.getMessage());
        }
    }

    synchronized int query(String query, ResultSet rs)
    throws SQLException {
        int code;

        try {
            out.write('Q'); /* Query */
            sendString(query);
            out.flush();

            if (rs != null) {
                synchronized (rs) {
                    gotARow = false;
                    do {
                        code = getAndProcessCommand(rs);
                    } while (code == QUERY_NOT_DONE);
                }
            } else {
                gotARow = false;
                do {
                    code = getAndProcessCommand(rs);
                } while (code == QUERY_NOT_DONE);
            }
        } catch (IOException e) {
            throw new SQLException("IO Error processing query: "
                                   + e.getMessage());
        }
        return code;
    }

    synchronized int query(String query)
    throws SQLException {
        return query(query, null);
    }

    synchronized void abortQuery(ResultSet rs)
    throws SQLException {
        Socket           sock;
        DataOutputStream out;

        /* open a new connection to the database, and send a cancel
         * request. */
        try {
            sock = new Socket(host, port);
            sock.setTcpNoDelay(true);
            /* make sure data is sent even if we close the socket. */
            sock.setSoLinger(true, 15);
            out = new DataOutputStream(sock.getOutputStream());
            out.writeInt(16);        /* size */
            out.writeInt(80877102);  /* cancel request code */
            out.writeInt(backendProcessId);
            out.writeInt(backendSecretKey);
            out.flush();
            out.close();
            sock.close();
//  System.err.println("sent a cancel request to the database");
        } catch (IOException e) {
            System.err.println("warning: unable to send abort message: "
                               + e.getMessage());
        }
        /* wait for the query to complete.  either it completes
         * normally, or the canel request works.  either way, we skip
         * everything up to the ready command. */
        if (!lastCodeWasReadyForQuery) {
//  System.err.println("will skip remaining parts of query.  telling you when i'm done...");
            try {
                if (rs != null) {
                    synchronized (rs) {
                        while (!getAndSkipCommand(rs))
                            ;
                    }
                } else {
                    while (!getAndSkipCommand(null))
                        ;
                }
            } catch (IOException e) {
                throw new SQLException("IO Error aborting query: "
                                       + e.getMessage());
            }
//  System.err.println("done skipping remaining parts of query.");
        }
    }

    synchronized void abortQuery()
    throws SQLException {
        abortQuery(null);
    }

    synchronized void close()
    throws SQLException {
        try {
            shutdown();
            closeSocket();
        } catch (IOException e) {
            throw new SQLException("Error closing database connection: "
                                   + e.getMessage());
        }
    }

    boolean isClosed()
    throws SQLException {
        return !databaseConnected;
    }
}
