/*------------------------------------------------------------------------
 |  FILE            ResultSet.java
 |  PART OF         Sverre's Simple JDBC Driver for PostgreSQL
 |  WRITTEN BY      Sverre H. Huseby <shh@thathost.com>
 +----------------------------------------------------------------------*/

package no.shhsoft.postgresql;

import java.util.*;
import java.io.*;
import java.sql.*;
import java.math.*;

/**
 * A ResultSet provides access to a table of data generated by
 * executing a Statement. The table rows are retrieved in
 * sequence. Within a row its column values can be accessed in any
 * order.
 * <P>
 * A ResultSet maintains a cursor pointing to its current row of data.
 * Initially the cursor is positioned before the first row.  The
 * 'next' method moves the cursor to the next row.
 * <P>
 * The getXXX methods retrieve column values for the current row.  You
 * can retrieve values either using the index number of the column, or
 * by using the name of the column.  In general using the column index
 * will be more efficient.  Columns are numbered from 1.
 * <P>
 * For maximum portability, ResultSet columns within each row should
 * be read in left-to-right order and each column should be read only
 * once.
 * <P>
 * For the getXXX methods, the JDBC driver attempts to convert the
 * underlying data to the specified Java type and returns a suitable
 * Java value.  See the JDBC specification for allowable mappings from
 * SQL types to Java types with the ResultSet.getXXX methods.
 * <P>
 * Column names used as input to getXXX methods are case insensitive.
 * When performing a getXXX using a column name, if several columns
 * have the same name, then the value of the first matching column
 * will be returned. The column name option is designed to be used
 * when column names are used in the SQL query. For columns that are
 * NOT explicitly named in the query, it is best to use column
 * numbers. If column names were used there is no way for the
 * programmer to guarantee that they actually refer to the intended
 * columns.
 * <P>
 * A ResultSet is automatically closed by the Statement that generated
 * it when that Statement is closed, re-executed, or is used to
 * retrieve the next result from a sequence of multiple results.
 * <P>
 * The number, types and properties of a ResultSet's columns are
 * provided by the ResulSetMetaData object returned by the getMetaData
 * method.
 *
 * @see      Statement#executeQuery
 * @see      Statement#getResultSet
 * @see      ResultSetMetaData
 * @author   Sverre H. Huseby
 *           &lt;<A HREF="mailto:shh@thathost.com">shh@thathost.com</A>&gt;
 * @version  $Id: ResultSet.java,v 1.10 2002/02/05 22:54:25 sverrehu Exp $
 */
public class ResultSet
implements java.sql.ResultSet {
/*-----------------------------------------------------------------------+
|  PRIVATE PART                                                          |
+-----------------------------------------------------------------------*/
    private boolean isClosed = false;

/*-----------------------------------------------------------------------+
|  PACKAGE LOCAL PART                                                    |
+-----------------------------------------------------------------------*/
    static final int DEFAULT_FETCH_SIZE = 64;

    int fetchSize;

    /* meta info.  filled in by PostgreSQLCommunicator.  used by
     * ResultSetMetaData. */
    /* arrays are counting from 0. */
    volatile int        numFields = -1;
    volatile String[]   fieldName;
    volatile Hashtable  fieldNameToIdx;  /* Integer, counting from 1 */
    volatile int[]      fieldTypeOID;
    volatile int[]      fieldTypeSize;
    volatile int[]      fieldTypeModifier;
    volatile int        notNullMapBytes;
    volatile byte[]     notNullMap;

    /* current fields.
     * indexes from left to right: preread row#, column#, value */
    volatile byte[][][] field;
    volatile int        numRows, currRow;

    /* the following is updated by
     * PostgreSQLCommunicator.getAndProcessCommand. */
    volatile boolean    fetchWithoutResult = false;


    ResultSet(Connection conn, Statement stmt)
    throws SQLException {
        this.conn = conn;
        this.stmt = stmt;

        setFetchSize(DEFAULT_FETCH_SIZE);

        numRows = 0;
        currRow = -1;
    }

    ResultSet()
    throws SQLException {
        this(null, null);
    }

/*-----------------------------------------------------------------------+
|  INHERITANTS' INTERFACE                                                |
+-----------------------------------------------------------------------*/
    protected Connection conn;
    protected Statement  stmt;
    protected int        lastColumnRead = -1; /* real array index */

    protected boolean isNull(int idx) { /* real array index */
        return (field[currRow][idx] == null);
    }

    /* counting from 1 */
    protected int getFieldIndexFromName(String name)
    throws SQLException {
        Integer i;

        if ((i = (Integer) fieldNameToIdx.get(name)) == null)
            throw new SQLException("no column named `" + name + "'");
        return i.intValue();
    }

    protected void assertOpen()
    throws SQLException {
        if (isClosed)
            throw new SQLException("ResultSet closed");
    }

    /* java.lang.Object *************************************************/
    protected void finalize() {
        try {
            close();
        } catch (SQLException e) {
        }
    }

/*-----------------------------------------------------------------------+
|  PUBLIC INTERFACE                                                      |
+-----------------------------------------------------------------------*/

    /* java.sql.ResultSet ***********************************************/
    /**
     * @exception  SQLException since the method is not implemented.
     */
    public synchronized boolean next()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * In some cases, it is desirable to immediately release a
     * ResultSet's database and JDBC resources instead of waiting for
     * this to happen when it is automatically closed; the close
     * method provides this immediate release.
     * <P>
     * <B>Note:</B> A ResultSet is automatically closed by the
     * Statement that generated it when that Statement is closed,
     * re-executed, or is used to retrieve the next result from a
     * sequence of multiple results. A ResultSet is also automatically
     * closed when it is garbage collected.
     *
     * @exception  SQLException if a database-access error occurs.
     */
    public synchronized void close()
    throws SQLException {
        if (stmt != null) {
            stmt.stopFetching();
            stmt = null;
        }
        isClosed = true;
    }

    /**
     * A column may have the value of SQL NULL; wasNull reports
     * whether the last column read had this special value.  Note that
     * you must first call getXXX on a column to try to read its value
     * and then call wasNull() to find if the value was the SQL NULL.
     *
     * @return     true if last column read was SQL NULL
     * @exception  SQLException if a database-access error occurs.
     */
    public boolean wasNull()
    throws SQLException {
        return isNull(lastColumnRead);
    }

    /**
     * Gets the value of a column in the current row as a Java String.
     *
     * @param      columnIndex
     *                     the first column is 1, the second is 2, ...
     * @return     the column value; if the value is SQL NULL, the
     *             result is null
     * @exception  SQLException if a database-access error occurs.
     */
    public String getString(int columnIndex)
    throws SQLException {
        byte[] value;

        lastColumnRead = columnIndex - 1;
        value = field[currRow][lastColumnRead];

        if (value == null)
            return null;
        return new String(value);
    }

    /**
     * Gets the value of a column in the current row as a Java
     * boolean.
     *
     * @param      columnIndex
     *                     the first column is 1, the second is 2, ...
     * @return     the column value; if the value is SQL NULL,
     *             the result is false
     * @exception  SQLException if a database-access error occurs.
     */
    public boolean getBoolean(int columnIndex)
    throws SQLException {
        String s;
        char   c;

        if ((s = getString(columnIndex)) == null)
            return false;
        return ((c = s.charAt(0)) == 't' || c == 'T');
    }

    /**
     * Gets the value of a column in the current row as a Java byte.
     *
     * @param      columnIndex
     *                     the first column is 1, the second is 2, ...
     * @return     the column value; if the value is SQL NULL,
     *             the result is 0
     * @exception  SQLException if a database-access error occurs.
     */
    public byte getByte(int columnIndex)
    throws SQLException {
        String s;

        if ((s = getString(columnIndex)) == null)
            return 0;
        try {
            return Byte.parseByte(s);
        } catch (NumberFormatException e) {
            throw new SQLException("Unable to getByte");
        }
    }

    /**
     * Gets the value of a column in the current row as a Java short.
     *
     * @param      columnIndex
     *                     the first column is 1, the second is 2, ...
     * @return     the column value; if the value is SQL NULL,
     *             the result is 0
     * @exception  SQLException if a database-access error occurs.
     */
    public short getShort(int columnIndex)
    throws SQLException {
        String s;

        if ((s = getString(columnIndex)) == null)
            return 0;
        try {
            return Short.parseShort(s);
        } catch (NumberFormatException e) {
            throw new SQLException("Unable to getShort");
        }
    }

    /**
     * Gets the value of a column in the current row as a Java int.
     *
     * @param      columnIndex
     *                     the first column is 1, the second is 2, ...
     * @return     the column value; if the value is SQL NULL,
     *             the result is 0
     * @exception  SQLException if a database-access error occurs.
     */
    public int getInt(int columnIndex)
    throws SQLException {
        String s;

        if ((s = getString(columnIndex)) == null)
            return 0;
        try {
            return Integer.parseInt(s);
        } catch (NumberFormatException e) {
            throw new SQLException("Unable to getInt");
        }
    }

    /**
     * Gets the value of a column in the current row as a Java long.
     *
     * @param      columnIndex
     *                     the first column is 1, the second is 2, ...
     * @return     the column value; if the value is SQL NULL,
     *             the result is 0
     * @exception  SQLException if a database-access error occurs.
     */
    public long getLong(int columnIndex)
    throws SQLException {
        String s;

        if ((s = getString(columnIndex)) == null)
            return 0L;
        try {
            return Long.parseLong(s);
        } catch (NumberFormatException e) {
            throw new SQLException("Unable to getLong");
        }
    }

    /**
     * Gets the value of a column in the current row as a Java float.
     *
     * @param      columnIndex
     *                     the first column is 1, the second is 2, ...
     * @return     the column value; if the value is SQL NULL,
     *             the result is 0
     * @exception  SQLException if a database-access error occurs.
     */
    public float getFloat(int columnIndex)
    throws SQLException {
        String s;

        if ((s = getString(columnIndex)) == null)
            return 0.0f;
        try {
            return Float.valueOf(s).floatValue();
        } catch (NumberFormatException e) {
            throw new SQLException("Unable to getFloat");
        }
    }

    /**
     * Gets the value of a column in the current row as a Java double.
     *
     * @param      columnIndex
     *                     the first column is 1, the second is 2, ...
     * @return     the column value; if the value is SQL NULL,
     *             the result is 0
     * @exception  SQLException if a database-access error occurs.
     */
    public double getDouble(int columnIndex)
    throws SQLException {
        String s;

        if ((s = getString(columnIndex)) == null)
            return 0.0;
        try {
            return Double.valueOf(s).doubleValue();
        } catch (NumberFormatException e) {
            throw new SQLException("Unable to getDouble");
        }
    }

    /**
     * Gets the value of a column in the current row as a
     * java.lang.BigDecimal object.
     *
     * @param      columnIndex
     *                     the first column is 1, the second is 2, ...
     * @param      scale   the number of digits to the right of the decimal
     * @return     the column value; if the value is SQL NULL,
     *             the result is null
     * @exception  SQLException if a database-access error occurs.
     */
    public BigDecimal getBigDecimal(int columnIndex, int scale)
    throws SQLException {
        String     s;
        BigDecimal ret;

        if ((s = getString(columnIndex)) == null)
            return null;
        try {
            ret = new BigDecimal(s);
        } catch (NumberFormatException e) {
            throw new SQLException("Unable to getBigDecimal");
        }
        try {
            ret.setScale(scale);
        } catch (ArithmeticException e) {
            throw new SQLException("Unable to getBigDecimal");
        }
        return ret;
    }

    /**
     * Gets the value of a column in the current row as a Java byte
     * array.  The bytes represent the raw values returned by the
     * driver.
     *
     * @param      columnIndex
     *                     the first column is 1, the second is 2, ...
     * @return     the column value; if the value is SQL NULL,
     *             the result is null
     * @exception  SQLException if a database-access error occurs.
     */
    public byte[] getBytes(int columnIndex)
    throws SQLException {
        lastColumnRead = columnIndex - 1;
        return field[currRow][lastColumnRead];
    }

    /**
     * Gets the value of a column in the current row as a
     * java.sql.Date object.
     *
     * @param      columnIndex
     *                     the first column is 1, the second is 2, ...
     * @return     the column value; if the value is SQL NULL,
     *             the result is null
     * @exception  SQLException if a database-access error occurs.
     */
    public java.sql.Date getDate(int columnIndex)
    throws SQLException {
        String s;

        if ((s = getString(columnIndex)) == null)
            return null;
        return java.sql.Date.valueOf(s);
    }

    /**
     * Gets the value of a column in the current row as a
     * java.sql.Time object.
     *
     * @param      columnIndex
     *                     the first column is 1, the second is 2, ...
     * @return     the column value; if the value is SQL NULL,
     *             the result is null
     * @exception  SQLException if a database-access error occurs.
     */
    public java.sql.Time getTime(int columnIndex)
    throws SQLException {
        String s;

        if ((s = getString(columnIndex)) == null)
            return null;
        return java.sql.Time.valueOf(s);
    }

    /**
     * Gets the value of a column in the current row as a
     * java.sql.Timestamp obje ct.
     *
     * @param      columnIndex
     *                     the first column is 1, the second is 2, ...
     * @return     the column value; if the value is SQL NULL,
     *             the result is null
     * @exception  SQLException if a database-access error occurs.
     */
    public java.sql.Timestamp getTimestamp(int columnIndex)
    throws SQLException {
        String s;

        if ((s = getString(columnIndex)) == null)
            return null;
        return java.sql.Timestamp.valueOf(s);
    }

    /**
     * A column value can be retrieved as a stream of ASCII characters
     * and then read in chunks from the stream.  This method is
     * particularly suitable for retrieving large LONGVARCHAR values.
     * The JDBC driver will do any necessary conversion from the
     * database format into ASCII.
     * <P>
     * <B>Note:</B> All the data in the returned stream must be read
     * prior to getting the value of any other column. The next call
     * to a get method implicitly closes the stream. . Also, a stream
     * may return 0 for available() whether there is data available or
     * not.
     *
     * @param      columnIndex
     *                     the first column is 1, the second is 2, ...
     * @return     a Java input stream that delivers the database column
     *             value as a stream of one byte ASCII characters.
     *             If the value is SQL NULL then the result is null.
     * @exception  SQLException if a database-access error occurs.
     */
    public java.io.InputStream getAsciiStream(int columnIndex)
    throws SQLException {
        /* TODO: should probably do some kind of ASCII mapping here. */
        return getBinaryStream(columnIndex);
    }

    /**
     * A column value can be retrieved as a stream of Unicode
     * characters and then read in chunks from the stream.  This
     * method is particularly suitable for retrieving large
     * LONGVARCHAR values.  The JDBC driver will do any necessary
     * conversion from the database format into Unicode.
     * <P>
     * <B>Note:</B> All the data in the returned stream must be read
     * prior to getting the value of any other column. The next call
     * to a get method implicitly closes the stream. . Also, a stream
     * may return 0 for available() whether there is data available or
     * not.
     *
     * @param      columnIndex
     *                     the first column is 1, the second is 2, ...
     * @return     a Java input stream that delivers the database column
     *             value as a stream of two byte Unicode characters.
     *             If the value is SQL NULL then the result is null.
     * @exception  SQLException if a database-access error occurs.
     */
    public java.io.InputStream getUnicodeStream(int columnIndex)
    throws SQLException {
        /* TODO: should probably do some kind of unicode mapping here. */
        return getBinaryStream(columnIndex);
    }

    /**
     * A column value can be retrieved as a stream of uninterpreted
     * bytes and then read in chunks from the stream.  This method is
     * particularly suitable for retrieving large LONGVARBINARY
     * values.
     * <P>
     * <B>Note:</B> All the data in the returned stream must be read
     * prior to getting the value of any other column. The next call
     * to a get method implicitly closes the stream. Also, a stream
     * may return 0 for available() whether there is data available or
     * not.
     *
     * @param      columnIndex
     *                     the first column is 1, the second is 2, ...
     * @return     a Java input stream that delivers the database column
     *             value as a stream of uninterpreted bytes.
     *             If the value is SQL NULL then the result is null.
     * @exception  SQLException if a database-access error occurs.
     */
    public java.io.InputStream getBinaryStream(int columnIndex)
    throws SQLException {
        byte b[];

        if ((b = getBytes(columnIndex)) == null)
            return null;
        return new ByteArrayInputStream(b);
    }

    /**
     * Gets the value of a column in the current row as a Java String.
     *
     * @param      columnName
     *                     is the SQL name of the column
     * @return     the column value; if the value is SQL NULL,
     *             the result is null
     * @exception  SQLException if a database-access error occurs.
     */
    public String getString(String columnName)
    throws SQLException {
        return getString(getFieldIndexFromName(columnName));
    }

    /**
     * Gets the value of a column in the current row as a Java
     * boolean.
     *
     * @param      columnName
     *                     is the SQL name of the column
     * @return     the column value; if the value is SQL NULL,
     *             the result is false
     * @exception  SQLException if a database-access error occurs.
     */
    public boolean getBoolean(String columnName)
    throws SQLException {
        return getBoolean(getFieldIndexFromName(columnName));
    }

    /**
     * Gets the value of a column in the current row as a Java byte.
     *
     * @param      columnName
     *                     is the SQL name of the column
     * @return     the column value; if the value is SQL NULL,
     *             the result is 0
     * @exception  SQLException if a database-access error occurs.
     */
    public byte getByte(String columnName)
    throws SQLException {
        return getByte(getFieldIndexFromName(columnName));
    }

    /**
     * Gets the value of a column in the current row as a Java short.
     *
     * @param      columnName
     *                     is the SQL name of the column
     * @return     the column value; if the value is SQL NULL,
     *             the result is 0
     * @exception  SQLException if a database-access error occurs.
     */
    public short getShort(String columnName)
    throws SQLException {
        return getShort(getFieldIndexFromName(columnName));
    }

    /**
     * Gets the value of a column in the current row as a Java int.
     *
     * @param      columnName
     *                     is the SQL name of the column
     * @return     the column value; if the value is SQL NULL,
     *             the result is 0
     * @exception  SQLException if a database-access error occurs.
     */
    public int getInt(String columnName)
    throws SQLException {
        return getInt(getFieldIndexFromName(columnName));
    }

    /**
     * Gets the value of a column in the current row as a Java long.
     *
     * @param      columnName
     *                     is the SQL name of the column
     * @return     the column value; if the value is SQL NULL,
     *             the result is 0
     * @exception  SQLException if a database-access error occurs.
     */
    public long getLong(String columnName)
    throws SQLException {
        return getLong(getFieldIndexFromName(columnName));
    }

    /**
     * Gets the value of a column in the current row as a Java float.
     *
     * @param      columnName
     *                     is the SQL name of the column
     * @return     the column value; if the value is SQL NULL,
     *             the result is 0
     * @exception  SQLException if a database-access error occurs.
     */
    public float getFloat(String columnName)
    throws SQLException {
        return getFloat(getFieldIndexFromName(columnName));
    }

    /**
     * Gets the value of a column in the current row as a Java double.
     *
     * @param      columnName
     *                     is the SQL name of the column
     * @return     the column value; if the value is SQL NULL,
     *             the result is 0
     * @exception  SQLException if a database-access error occurs.
     */
    public double getDouble(String columnName)
    throws SQLException {
        return getDouble(getFieldIndexFromName(columnName));
    }

    /**
     * Gets the value of a column in the current row as a
     * java.lang.BigDecimal ob ject.
     *
     * @param      columnName
     *                     is the SQL name of the column
     * @param      scale   the number of digits to the right of the decimal
     * @return     the column value; if the value is SQL NULL,
     *             the result is null
     * @exception  SQLException if a database-access error occurs.
     */
    public BigDecimal getBigDecimal(String columnName, int scale)
    throws SQLException {
        return getBigDecimal(getFieldIndexFromName(columnName), scale);
    }

    /**
     * Gets the value of a column in the current row as a Java byte
     * array.  The bytes represent the raw values returned by the
     * driver.
     *
     * @param      columnName
     *                     is the SQL name of the column
     * @return     the column value; if the value is SQL NULL,
     *             the result is null
     * @exception  SQLException if a database-access error occurs.
     */
    public byte[] getBytes(String columnName)
    throws SQLException {
        return getBytes(getFieldIndexFromName(columnName));
    }

    /**
     * Gets the value of a column in the current row as a
     * java.sql.Date object.
     *
     * @param      columnName
     *                     is the SQL name of the column
     * @return     the column value; if the value is SQL NULL,
     *             the result is null
     * @exception  SQLException if a database-access error occurs.
     */
    public java.sql.Date getDate(String columnName)
    throws SQLException {
        return getDate(getFieldIndexFromName(columnName));
    }

    /**
     * Gets the value of a column in the current row as a
     * java.sql.Time object.
     *
     * @param      columnName
     *                     is the SQL name of the column
     * @return     the column value; if the value is SQL NULL,
     *             the result is null
     * @exception  SQLException if a database-access error occurs.
     */
    public java.sql.Time getTime(String columnName)
    throws SQLException {
        return getTime(getFieldIndexFromName(columnName));
    }

    /**
     * Gets the value of a column in the current row as a
     * java.sql.Timestamp obje ct.
     *
     * @param      columnName
     *                     is the SQL name of the column
     * @return     the column value; if the value is SQL NULL,
     *             the result is null
     * @exception  SQLException if a database-access error occurs.
     */
    public java.sql.Timestamp getTimestamp(String columnName)
    throws SQLException {
        return getTimestamp(getFieldIndexFromName(columnName));
    }

    /**
     * A column value can be retrieved as a stream of ASCII characters
     * and then read in chunks from the stream.  This method is
     * particularly suitable for retrieving large LONGVARCHAR values.
     * The JDBC driver will do any necessary conversion from the
     * database format into ASCII.
     * <P>
     * <B>Note:</B> All the data in the returned stream must be read
     * prior to getting the value of any other column. The next call
     * to a get method implicitly closes the stream.
     *
     * @param      columnName
     *                     is the SQL name of the column
     * @return     a Java input stream that delivers the database column
     *             value as a stream of one byte ASCII characters.
     *             If the value is SQL NULL then the result is null.
     * @exception  SQLException if a database-access error occurs.
     */
    public java.io.InputStream getAsciiStream(String columnName)
    throws SQLException {
        return getAsciiStream(getFieldIndexFromName(columnName));
    }

    /**
     * A column value can be retrieved as a stream of Unicode
     * characters and then read in chunks from the stream.  This
     * method is particularly suitable for retrieving large
     * LONGVARCHAR values.  The JDBC driver will do any necessary
     * conversion from the database format into Unicode.
     * <P>
     * <B>Note:</B> All the data in the returned stream must be read
     * prior to getting the value of any other column. The next call
     * to a get method implicitly closes the stream.
     *
     * @param      columnName
     *                     is the SQL name of the column
     * @return     a Java input stream that delivers the database column
     *             value as a stream of two byte Unicode characters.
     *             If the value is SQL NULL then the result is null.
     * @exception  SQLException if a database-access error occurs.
     */
    public java.io.InputStream getUnicodeStream(String columnName)
    throws SQLException {
        return getUnicodeStream(getFieldIndexFromName(columnName));
    }

    /**
     * A column value can be retrieved as a stream of uninterpreted
     * bytes and then read in chunks from the stream.  This method is
     * particularly suitable for retrieving large LONGVARBINARY
     * values.
     * <P>
     * <B>Note:</B> All the data in the returned stream must be read
     * prior to getting the value of any other column. The next call
     * to a get method implicitly closes the stream.
     *
     * @param      columnName
     *                     is the SQL name of the column
     * @return     a Java input stream that delivers the database column
     *             value as a stream of uninterpreted bytes.
     *             If the value is SQL NULL then the result is null.
     * @exception  SQLException if a database-access error occurs.
     */
    public java.io.InputStream getBinaryStream(String columnName)
    throws SQLException {
        return getBinaryStream(getFieldIndexFromName(columnName));
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public SQLWarning getWarnings()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void clearWarnings()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * Gets the name of the SQL cursor used by this ResultSet.
     * <P>
     * In SQL, a result table is retrieved through a cursor that is
     * named. The current row of a result can be updated or deleted
     * using a positioned update/delete statement that references the
     * cursor name.
     * <P>
     * JDBC supports this SQL feature by providing the name of the SQL
     * cursor used by a ResultSet. The current row of a ResultSet is
     * also the current row of this SQL cursor.
     * <P>
     * <B>Note:</B> If positioned update is not supported a
     * SQLException is thrown
     *
     * @return     the ResultSet's SQL cursor name
     * @exception  SQLException if a database-access error occurs.
     */
    public String getCursorName()
    throws SQLException {
        throw new SQLException("Positioned update not supported.");
    }

    /**
     * The number, types and properties of a ResultSet's columns are
     * provided by the getMetaData method.
     *
     * @return     the description of a ResultSet's columns
     * @exception  SQLException if a database-access error occurs.
     */
    public java.sql.ResultSetMetaData getMetaData()
    throws SQLException {
        return new ResultSetMetaData(this, conn);
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public Object getObject(int columnIndex)
    throws SQLException {
        lastColumnRead = columnIndex - 1;
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * Gets the value of a column in the current row as a Java object.
     * <P>
     * This method will return the value of the given column as a Java
     * object.  The type of the Java object will be the default Java
     * Object type corresponding to the column's SQL type, following
     * the mapping specified in the JDBC spec.
     * <P>
     * This method may also be used to read datatabase specific
     * abstract data types.
     *
     * @param      columnName
     *                     is the SQL name of the column
     * @return     A java.lang.Object holding the column value.
     * @exception  SQLException if a database-access error occurs.
     */
    public Object getObject(String columnName)
    throws SQLException {
        return getObject(getFieldIndexFromName(columnName));
    }

    /**
     * Maps a Resultset column name to a ResultSet column index.
     *
     * @param      columnName
     *                     the name of the column
     * @return     the column index
     * @exception  SQLException if a database-access error occurs.
     */
    public int findColumn(String columnName)
    throws SQLException {
        return getFieldIndexFromName(columnName);
    }

    /* -- JDBC 2.0 ---------------------------------------------------- */

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.io.Reader getCharacterStream(int columnIndex)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.io.Reader getCharacterStream(String columnName)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public BigDecimal getBigDecimal(int columnIndex)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public BigDecimal getBigDecimal(String columnName)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean isBeforeFirst()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean isAfterLast()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean isFirst()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean isLast()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void beforeFirst()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void afterLast()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean first()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean last()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public int getRow()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean absolute(int row)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean relative(int rows)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean previous()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * The constant indicating that the rows in a result set will be
     * processed in a forward direction; first-to-last.  This constant
     * is used by the method <code>setFetchDirection</code> as a hint
     * to the driver, which the driver may ignore.
     * @since 1.2
     * @see <a href="package-summary.html#2.0 API">What Is in the JDBC
     *      2.0 API</a>
     */
    public int FETCH_FORWARD = 1000;

    /**
     * The constant indicating that the rows in a result set will be
     * processed in a reverse direction; last-to-first.  This constant
     * is used by the method <code>setFetchDirection</code> as a hint
     * to the driver, which the driver may ignore.
     * @since 1.2
     * @see <a href="package-summary.html#2.0 API">What Is in the JDBC
     *      2.0 API</a>
     */
    public int FETCH_REVERSE = 1001;

    /**
     * The constant indicating that the order in which rows in a
     * result set will be processed is unknown.  This constant is used
     * by the method <code>setFetchDirection</code> as a hint to the
     * driver, which the driver may ignore.
     */
    public int FETCH_UNKNOWN = 1002;

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void setFetchDirection(int direction)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public int getFetchDirection()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * Gives the JDBC driver a hint as to the number of rows that
     * should be fetched from the database when more rows are needed
     * for this <code>ResultSet</code> object.  If the fetch size
     * specified is zero, the JDBC driver ignores the value and is
     * free to make its own best guess as to what the fetch size
     * should be.  The default value is set by the
     * <code>Statement</code> object that created the result set.  The
     * fetch size may be changed at any time.
     *
     * @param      rows the number of rows to fetch
     * @exception  SQLException if a database access error occurs or
     *             the condition <code>0 <= rows <=
     *             this.getMaxRows()</code> is not satisfied
     * @since      1.2
     * @see        <a href="package-summary.html#2.0 API">What Is in
     *             the JDBC 2.0 API</a>
     */
    public synchronized void setFetchSize(int rows)
    throws SQLException {
        byte[][][] newField;
        int        rowIdx, colIdx, valIdx;

        /* cannot decrease the size of the field buffer, if we
         * have already read something into it. */
        if (numRows <= rows) {
            newField = new byte[rows][][];
            for (rowIdx = numRows - 1; rowIdx >= 0; rowIdx--)
                for (colIdx = field[rowIdx].length - 1; colIdx >= 0; colIdx--)
                    for (valIdx = field[rowIdx][colIdx].length - 1;
                         valIdx >= 0; valIdx--)
                        newField[rowIdx][colIdx][valIdx]
                            = field[rowIdx][colIdx][valIdx];
            field = newField;
        }

        fetchSize = rows;
    }

    /**
     * Returns the fetch size for this <code>ResultSet</code> object.
     *
     * @return     the current fetch size for this <code>ResultSet</code>
     *             object
     * @exception  SQLException if a database access error occurs
     * @since      1.2
     * @see        <a href="package-summary.html#2.0 API">What Is in
     *             the JDBC 2.0 API</a>
     */
    public int getFetchSize()
    throws SQLException {
        return fetchSize;
    }

    /**
     * The constant indicating the type for a <code>ResultSet</code>
     * object whose cursor may move only forward.
     * @since 1.2
     * @see <a href="package-summary.html#2.0 API">What Is in the JDBC
     *      2.0 API</a>
     */
    public int TYPE_FORWARD_ONLY = 1003;

    /**
     * The constant indicating the type for a <code>ResultSet</code>
     * object that is scrollable but generally not sensitive to
     * changes made by others.
     * @since 1.2
     * @see <a href="package-summary.html#2.0 API">What Is in the JDBC
     *      2.0 API</a>
         *
     */
    public int TYPE_SCROLL_INSENSITIVE = 1004;

    /**
     * The constant indicating the type for a <code>ResultSet</code>
     * object that is scrollable and generally sensitive to changes
     * made by others.
     * @since 1.2
     * @see <a href="package-summary.html#2.0 API">What Is in the JDBC
     *      2.0 API</a>
     */
    public int TYPE_SCROLL_SENSITIVE = 1005;

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public int getType()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * The constant indicating the concurrency mode for a
     * <code>ResultSet</code> object that may NOT be updated.
     * @since 1.2
     * @see <a href="package-summary.html#2.0 API">What Is in the JDBC
     *      2.0 API</a>
     *
     */
    public int CONCUR_READ_ONLY = 1007;

    /**
     * The constant indicating the concurrency mode for a
     * <code>ResultSet</code> object that may be updated.
     * @since 1.2
     * @see <a href="package-summary.html#2.0 API">What Is in the JDBC
     *      2.0 API</a>
     *
     */
    public int CONCUR_UPDATABLE = 1008;

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public int getConcurrency()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean rowUpdated()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean rowInserted()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean rowDeleted()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateNull(int columnIndex)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateBoolean(int columnIndex, boolean x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateByte(int columnIndex, byte x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateShort(int columnIndex, short x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateInt(int columnIndex, int x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateLong(int columnIndex, long x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateFloat(int columnIndex, float x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateDouble(int columnIndex, double x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateBigDecimal(int columnIndex, BigDecimal x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateString(int columnIndex, String x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateBytes(int columnIndex, byte x[])
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateDate(int columnIndex, java.sql.Date x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateTime(int columnIndex, java.sql.Time x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateTimestamp(int columnIndex, java.sql.Timestamp x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateAsciiStream(int columnIndex, java.io.InputStream x,
                                  int length)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateBinaryStream(int columnIndex, java.io.InputStream x,
                                   int length)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateCharacterStream(int columnIndex, java.io.Reader x,
                                      int length)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateObject(int columnIndex, Object x, int scale)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateObject(int columnIndex, Object x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateNull(String columnName)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateBoolean(String columnName, boolean x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateByte(String columnName, byte x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateShort(String columnName, short x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateInt(String columnName, int x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateLong(String columnName, long x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateFloat(String columnName, float x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateDouble(String columnName, double x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateBigDecimal(String columnName, BigDecimal x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateString(String columnName, String x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateBytes(String columnName, byte x[])
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateDate(String columnName, java.sql.Date x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateTime(String columnName, java.sql.Time x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateTimestamp(String columnName, java.sql.Timestamp x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateAsciiStream(String columnName, java.io.InputStream x,
                                  int length)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateBinaryStream(String columnName, java.io.InputStream x,
                                   int length)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateCharacterStream(String columnName, java.io.Reader reader,
                                      int length)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateObject(String columnName, Object x, int scale)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateObject(String columnName, Object x)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void insertRow()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void updateRow()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void deleteRow()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void refreshRow()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void cancelRowUpdates()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void moveToInsertRow()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void moveToCurrentRow()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.Statement getStatement()
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public Object getObject(int i, java.util.Map map)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public Ref getRef(int i)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public Blob getBlob(int i)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public Clob getClob(int i)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public Array getArray(int i)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public Object getObject(String colName, java.util.Map map)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public Ref getRef(String colName)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public Blob getBlob(String colName)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public Clob getClob(String colName)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public Array getArray(String colName)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.Date getDate(int columnIndex, Calendar cal)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.Date getDate(String columnName, Calendar cal)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.Time getTime(int columnIndex, Calendar cal)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.Time getTime(String columnName, Calendar cal)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.Timestamp getTimestamp(int columnIndex, Calendar cal)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.Timestamp getTimestamp(String columnName, Calendar cal)
    throws SQLException {
        assertOpen();
        throw new NotImplementedException();
        /* TODO */
    }
}
