/*------------------------------------------------------------------------
 |  FILE            JDBCSpeed.java
 |  PART OF
 |  WRITTEN BY      Sverre H. Huseby <shh@thathost.com>
 +----------------------------------------------------------------------*/

package test;

import java.util.*;
import java.io.*;

import java.net.URL;
import java.sql.*;

/**
 * Speed ratio between original JDBC driver and my driver running a
 * simple query from a table containing a few hundred records.  Both
 * drivers were compiled with optimization.  Times may vary according
 * to VM, computer, load etc., so you should look at the ratio only.
 * Both drivers are compiled with the same optimalization settings.
 * The original driver is the latest driver on the time given.  Median
 * values are used for calculations.
 *
 * Date and time        Original     Mine        Ratio     NewOrig
 *
 * 2000-02-25 23:49        580        191        3.037        x
 * 2000-02-27 23:58        518        194        2.670
 * 2000-04-20 23:13        549        188        2.920        x  (PG 6.5.3)
 * 2000-06-29 22:38        279        187        1.492        x  (PG 7.0.2)
 * 2002-02-05 23:59         54         21        2.571        x  (PG 7.2.0)
 *
 * (2002-02-05: I got new hardware.  lucky me.)
 *
 * ------------------------------------------------------------------------
 *
 * Memory usage ratio with the same settings described above.  Values are
 * median, as output by this program.  Please note that measuring memory
 * usage in a Java program is not by far an exact science!
 *
 * Date and time        Original     Mine        Ratio     NewOrig
 * 2000-06-29 22:38     1483400     285004        5.2         x  (PG 7.0.2)
 * 2002-02-06 23:59     1485568     305440        4.9         x  (PG 7.2.0)
 *
 * ------------------------------------------------------------------------
 *
 * Old results (don't bother trying to understand this):
 *
 * PostgreSQL STANDARD DRIVER:
 *                            -O:      avg: 1380 ms, min: 1332 ms
 *                       + final:      avg: 1367 ms, min: 1309 ms
 *  + minor changes in PG_Stream:      avg: 1340 ms, min: 1285 ms
 *             + Str.toLowerCase:      avg: 1553 ms, min: 1504 ms (!)
 *             - Str.toLowerCase
 *         + BufferedInputStream:      avg:  563 ms, min;  537 ms
 *
 * MY DRIVER:
 *                              :      avg: 1148 ms, min: 1113 ms
 *            prereading 32 rows:      avg:  222 ms, min:  203 ms (whee!)
 *
 * ------------------------------------------------------------------------
 *
 * @author   Sverre H. Huseby
 *           &lt;<A HREF="mailto:shh@thathost.com">shh@thathost.com</A>&gt;
 * @version  $Id: JDBCSpeed.java,v 1.8 2002/02/05 23:19:40 sverrehu Exp $
 */
public final class JDBCSpeed
extends TestCommon {
/*-----------------------------------------------------------------------+
 |  PRIVATE PART                                                         |
 +----------------------------------------------------------------------*/
    private static final int NUM_TESTS = 200;
    private static String driver, url;

    private void runTest(String driver) {
        int   q;
        long  t, tTot = 0, tMin = Long.MAX_VALUE, tMax = Long.MIN_VALUE;
        long  m, mTot = 0, mMin = Long.MAX_VALUE, mMax = Long.MIN_VALUE;
        int   n;
        long[] times;
        long[] mems;

        System.out.println("Testing driver `" + driver + "'");
        times = new long[NUM_TESTS];
        mems = new long[NUM_TESTS];
        for (q = 0; q < NUM_TESTS; q++) {
            fullGarbageCollection();
            m = Runtime.getRuntime().freeMemory();
            t = System.currentTimeMillis();
            n = execute("SELECT * FROM Question");
            t = System.currentTimeMillis() - t;
            m = m - Runtime.getRuntime().freeMemory();
            System.out.print("\r" + (q + 1) + "/" + NUM_TESTS + ": "
                             + n + " rows returned in " + t + " ms "
                             + "using " + m + " bytes    ");
            tTot += t;
            if (t < tMin)
                tMin = t;
            if (t > tMax)
                tMax = t;
            times[q] = t;
            mTot += m;
            if (m < mMin)
                mMin = m;
            if (m > mMax)
                mMax = m;
            mems[q] = m;
        }
        System.out.print("\r                                          "
                         + "                           \r");
        System.out.println("average time: " + (tTot / q) + " ms");
        System.out.println("median  time: " + getMedian(times) + " ms");
        System.out.println("minimum time: " + tMin);
        System.out.println("maximum time: " + tMax);

        System.out.println("average mem: " + (mTot / q) + " bytes");
        System.out.println("median  mem: " + getMedian(mems) + " bytes");
        System.out.println("minimum mem: " + mMin);
        System.out.println("maximum mem: " + mMax);
    }

/*-----------------------------------------------------------------------+
 |  PUBLIC INTERFACE                                                     |
 +----------------------------------------------------------------------*/
    public static void main(String[] args) {
        JDBCSpeed jdbcSpeed;

        if (args.length != 2) {
            System.err.println("arguments: driver url");
            System.exit(1);
        }
        driver = args[0];
        url = args[1];
        jdbcSpeed = new JDBCSpeed();
        jdbcSpeed.connect(driver, url);
        jdbcSpeed.runTest(driver);
        jdbcSpeed.close();
    }
}
