/*------------------------------------------------------------------------
 |  FILE            PostgreSQL.java
 |  PART OF
 |  WRITTEN BY      Sverre H. Huseby <shh@thathost.com>
 +----------------------------------------------------------------------*/

package test;

import java.util.*;
import java.net.*;
import java.io.*;

/**
 *
 *
 * @author   Sverre H. Huseby
 *           &lt;<A HREF="mailto:shh@thathost.com">shh@thathost.com</A>&gt;
 * @version  $Id: PostgreSQL.java,v 1.3 2002/02/05 23:08:16 sverrehu Exp $
 */
public class PostgreSQL {
/*-----------------------------------------------------------------------+
|  PRIVATE PART                                                          |
+-----------------------------------------------------------------------*/
    Socket sock;
    BufferedOutputStream out;
    BufferedInputStream  in;

    private void sendInt32(int i)
    throws IOException {
        out.write((byte) ((i >> 24) & 0xFF));
        out.write((byte) ((i >> 16) & 0xFF));
        out.write((byte) ((i >>  8) & 0xFF));
        out.write((byte) ( i        & 0xFF));
    }

    private int receiveInt32()
    throws IOException {
        int ret;

        ret = in.read();
        ret = (ret << 8) | in.read();
        ret = (ret << 8) | in.read();
        ret = (ret << 8) | in.read();
        return ret;
    }

    private void sendInt16(int i)
    throws IOException {
        out.write((byte) ((i >>  8) & 0xFF));
        out.write((byte) ( i        & 0xFF));
    }

    private int receiveInt16()
    throws IOException {
        int ret;

        ret = in.read();
        ret = (ret << 8) | in.read();
        return ret;
    }

    private void sendLimString(int n, String s)
    throws IOException {
        byte[] buff;
        int    slen;

        buff = new byte[n];
        slen = s.length();
        if (slen > n)
            slen = n;
        else
            buff[n - 1] = 0;
        s.getBytes(0, slen, buff, 0);
        out.write(buff);
    }

    private void sendString(String s)
    throws IOException {
        out.write(s.getBytes());
        out.write('\0');
    }

    private String receiveString()
    throws IOException {
        StringBuffer sb;
        int          c;

        sb = new StringBuffer();
        while ((c = in.read()) > 0)
            sb.append((char) c);
        return sb.toString();
    }

    private void receiveByteN(int n, byte[] buff)
    throws IOException {
        in.read(buff, 0, n);
    }

    private int getCommand()
    throws IOException {
        int    code;
        String msg;

        for (;;) {
            code = in.read();
            if (code == 'E') {
                /* ErrorResponse */
                msg = receiveString();
                throw new IOException(msg);
            } else if (code == 'N') {
                /* NoticeResponse */
                msg = receiveString();
                System.err.println("WARNING: " + msg);
            } else
                break;
        }
        return code;
    }

    private void startup(String database, String username, String password)
    throws IOException {
        int     code, subCode;
        boolean authOK = false;
        boolean backendReady = false;
        String  msg;

        /* talk to the postmaster to log in */
        sendInt32(296); /* length */
        sendInt32((2 << 16) | 0);  /* protocol version */
        sendLimString(64, database);
        sendLimString(32, username);
        sendLimString(64, "");
        sendLimString(64, "");
        sendLimString(64, "");
        out.flush();

        do {
            try {
                code = getCommand();
            } catch (IOException e) {
                sock.close();
                throw e;
            }
            switch (code) {
              case 'R': /* Authentication... */
                subCode = receiveInt32();
                switch (subCode) {
                  case 0: /* AuthenticationOK */
                    authOK = true;
                    break;
                  case 1: /* AuthenticationKerberosV4 */
                  case 2: /* AuthenticationKerberosV5 */
                    sock.close();
                    throw new IOException(
                               "Unsupported authentication method required");
                  case 3: /* AuthenticationUnencryptedPassword */
                    sendInt32(4 + password.length() + 1);
                    sendString(password);
                    out.flush();
                    break;
                  case 4: /* AuthenticationEncryptedPassword */
                    in.read(); /* skip salt byte one */
                    in.read(); /* skip salt byte two */
                    sock.close();
                    throw new IOException(
                               "Unsupported authentication method required");
                  default:
                    sock.close();
                    throw new IOException(
                               "Unknown authentication method required");
                }
                break;
              default:
                sock.close();
                throw new IOException("Unknown response from backend");
            }
        } while (!authOK);

        /* wait for backend to wake up */
        do {
            try {
                code = getCommand();
            } catch (IOException e) {
                sock.close();
                throw e;
            }
            switch (code) {
              case 'K': /* BackendKeyData */
                receiveInt32(); /* backend process ID */
                receiveInt32(); /* backend secret key */
                break;
              case 'Z': /* ReadyForQuery */
                backendReady = true;
                break;
              default:
                sock.close();
                throw new IOException("Unknown response from backend");
            }
        } while (!backendReady);
    }

    private void connect(String host, int port, String database,
                         String username, String password)
    throws IOException {
        sock = new Socket(host, port);
        sock.setTcpNoDelay(true);
        out = new BufferedOutputStream(sock.getOutputStream());
        in = new BufferedInputStream(sock.getInputStream());

        startup(database, username, password);
    }

    private void disconnect()
    throws IOException {
        out.write('X'); /* Terminate */
        out.flush();
        sock.close();
    }

    private synchronized void query(String query)
    throws IOException {
        int     code, q;
        boolean queryDone = false;
        String  s;

        /* field information */
        int numFields = -1;
        String[] fieldName = null;
        int[]    fieldOID = null;
        int[]    fieldTypeSize = null;
        int[]    fieldTypeModifier = null;
        int      notNullMapBytes = 0;
        byte[]   notNullMap = null;

        out.write('Q'); /* Query */
        sendString(query);
        out.flush();

        do {
            code = getCommand();
            switch (code) {
              case 'B': /* BinaryRow */
                System.err.println("BinaryRow");
                /* TODO */
                break;
              case 'C': /* CompletedResponse */
                /* an SQL command (a query may contain several
                 * commands) completed normally. */
                s = receiveString(); /* tag */
                System.err.println("CompletedResponse: " + s);
                break;
              case 'D': /* AsciiRow */
                System.err.println("AsciiRow");
                if (numFields < 0)
                    throw new IOException("Unexpected AsciiRow");
                receiveByteN(notNullMapBytes, notNullMap);
                int maskBit = 0x80;
                int mapByte = 0;
                int size;
                for (q = 0; q < numFields; q++) {
                    if ((notNullMap[mapByte] & maskBit) != 0) {
                        /* not null field */
                        System.err.println("not null: " + fieldName[q]);
                        size = receiveInt32();
                        in.skip(size - 4);
                    } else {
                        /* null field */
                        System.err.println("    null: " + fieldName[q]);
                    }
                    if ((maskBit >>= 1) == 0) {
                        maskBit = 0x80;
                        ++mapByte;
                    }
                }
                /* TODO */
                break;
              case 'G': /* CopyInResponse */
                System.err.println("CopyInResponse");
                break;
              case 'H': /* CopyOutResponse */
                System.err.println("CopyOutResponse");
                /* TODO: a CopyDataRows message will follow. */
                break;
              case 'I': /* EmptyQuery */
                receiveString(); /* unused */
                System.err.println("EmptyQuery");
                break;
              case 'P': /* CursorResponse */
                s = receiveString(); /* cursor name */
                System.err.println("CursorResponse: " + s);
                /* a CompletedResponse or RowDescription will follow. */
                break;
              case 'T': /* RowDescription */
                System.err.println("RowDescription");
                numFields = receiveInt16();
                fieldName = new String[numFields];
                fieldOID = new int[numFields];
                fieldTypeSize = new int[numFields];
                fieldTypeModifier = new int[numFields];

                for (q = 0; q < numFields; q++) {
                    fieldName[q] = receiveString();
                    fieldOID[q] = receiveInt32();
                    fieldTypeSize[q] = receiveInt16(); /* type size */
                    fieldTypeModifier[q] = receiveInt32(); /* type modifier */
                }
                if (numFields > 0) {
                    notNullMapBytes = ((numFields - 1) >> 3) + 1;
                    notNullMap = new byte[notNullMapBytes];
                } else {
                    notNullMapBytes = 0;
                }
                /* AsciiRow or BinaryRow will follow. */
                break;
              case 'Z': /* ReadyForQuery */
                /* received when everything is finished, regardless of
                 * errors. */
                System.err.println("ReadyForQuery");
                queryDone = true;
                break;
            }
        } while (!queryDone);
    }

    private void commandLoop() {
        String         cmd;
        BufferedReader in;

        in = new BufferedReader(new InputStreamReader(System.in));
        for (;;) {
            System.out.print("JDBC> ");
            try {
                cmd = in.readLine().trim();
            } catch (IOException e) {
                System.err.println(e.getMessage());
                break;
            }
            if (cmd.equalsIgnoreCase("QUIT") || cmd.equalsIgnoreCase("EXIT"))
                break;
            try {
                query(cmd);
            } catch (Exception e) {
                System.err.println("exception: " + e.getMessage());
                e.printStackTrace(System.err);
            }
        }
    }

    private void run()
    throws IOException {
        connect("localhost", 5432, "quiz", "sverrehu", "3142");
        query("SET DATESTYLE TO 'ISO'");
        commandLoop();
        disconnect();
    }

/*-----------------------------------------------------------------------+
|  PUBLIC INTERFACE                                                      |
+-----------------------------------------------------------------------*/
    public static void main(String[] args) {
        try {
            (new PostgreSQL()).run();
        } catch (Exception e) {
            System.err.println("exception: " + e.getMessage());
            e.printStackTrace(System.err);
        }
    }
}
